#!/usr/bin/perl
#
# Bitflu - (C) 2006 Adrian Ulrich
#
# Released under the terms of The "Artistic License 2.0".
# http://www.perlfoundation.org/legal/licenses/artistic-2_0.txt
#
#use warnings;
use strict;



## Feature: Selektives downloaden
## -> idee: Nicht-Bentigte pieces berechnen und in .free umbenennen: .free/0 -> .free/_MUTED_0
##    .. muss alles mit _getfreedir gereviewt werden (SuggestPiece / VerifyStorageDir mssen definitiv gefixt/gendert werden)
## -> Commit muss auch abgendert werden


###
### === TODO ===
###
###  * No Endgame
###  * Periodic Tracker-tracking not (really) implemented
###  * XXX_UGLY: Dynamic loop could need some cleanups : But it seems to work okay so far..
###  * Upload Shaping (BW_STATS) is -not-so-good-
###  * Unchoking / Ranking : Is it fair? I guess not..
###  * Erfolgreicher commit (!xcommit) knnte .done/0..GetNumberOfPieces unlinken
use IO::Socket;
use IO::Select;
use Digest::SHA1;
use Data::Dumper;
use Getopt::Long;
use List::Util 'shuffle';
use Carp qw(confess croak cluck carp);
use URI::Escape;
use Scalar::Util; #For chroot
use File::Find;
use POSIX;



use constant BITFLU_BUILD   => '6110';  # YMMB(uild)
use constant BITFLU_VERSION => '0.10';

$SIG{PIPE} = $SIG{CHLD} = 'IGNORE';

$SIG{INT} = $SIG{HUP} = \&ConfDown;


## LogLevel constants. Everything above 'LOGLEVEL' will
## be printed to STDOUT.
use constant L_DEBUG       => 001;  # LowLevel Debug
use constant L_DEBUG_STORE => 002;  # Storemanager
use constant L_DEBUG_RPICE => 003;  # ReadPiece
use constant L_DEBUG_SUGST => 004;  # SuggestNewPiece
use constant L_DEBUG_CRX   => 005;  # CreateRequest
use constant L_DEBUG_PBLSH => 006;  # Publish
use constant L_DEBUG_CONFIG=> 899;  # dumps config
use constant L_INFO        => 100;  # Info messages
use constant L_WARN        => 200;  # Warning messages
use constant L_PANIC       => 900;  # Panics client
use constant LOGLEVEL      => 199;

## BitTorrentState constants. Do not change unless you know what
## you are doing.
use constant BTS_WRITE_HANDSHAKE          => 100; # -> Switches to BTS_READ_HANDSHAKE_RESPONSE
use constant BTS_READ_HANDSHAKE_RESPONSE  => 101; # -> Switches to BTS_WRITE_BITFIELD
use constant BTS_WRITE_BITFIELD           => 200; # -> Switches to BTS_IDLE
use constant BTS_IDLE                     => 300; # -> MAIN()
use constant BTS_CONNECTING               => 400; # -> Switches to BTS_WRITE_HANDSHAKE
use constant BTS_READ_CLIENT_HANDSHAKE    => 500;
use constant BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD => 600;
use constant BTS_KILLME_PLEASE            => 998; # -> Request to close the connection on next bts_writer run
use constant BTS_UNDEF                    => 999; # -> Undef state
## State 'flow':
#### If *WE* connect to a client
# -> BTS_CONNECTING -> BTS_WRITE_HANDSHAKE -> BTS_READ_HANDSHAKE_RESPONSE -> BTS_WRITE_BITFIELD -> BTS_IDLE
#
### If a client connects to US
# -> BTS_READ_CLIENT_HANDSHAKE -> BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD -> BTS_WRITE_BITFIELD -> BTS_IDLE


## Dynamic select loop
## we are running with DYNASLEEP_MAX but are going
## to switch to DYNASLEEP_MIN if we need to (Fast download)
## Change this to reduce CPU-Usage and/or improve download speed
use constant DYNASLEEP_MIN =>0.00001;
use constant DYNASLEEP_MAX =>0.10; #0.12

## Piece Size; do not change unless you know what you are doing!
use constant PIECESIZE    => (2**14); #Default size for pieces
use constant MAXPIECESIZE => (2**17); #Drop the connection if a client requests more..

## Various Timeout-Values.
use constant TIMEOUT_SHAKE             => 16;     # Small timeout for things like 'connecting' and handshaking / bitfield exchange                 {TIMEOUT}
use constant TIMEOUT_INTERESTED        => 180;    # Send an INTERESTED request each X+rand(X) seconds if client has a good piece for us
use constant TIMEOUT_INTERESTED_MAX    => 2500;   # Do not go above TIMEOUT_INTERESTED
use constant TIMEOUT_WAITFORPIECE      => 60*3;   # Wait X seconds for a piece to arrive                                                           {TIMEOUT_WAITFORPIECE}
use constant TIMEOUT_MECHOKED          => 60*120; # Drop connection if we are choked for more than X seconds                                       {TIMEOUT_MECHOKED}
use constant TIMEOUT_DOUBLECHOKE_DRIFT => 60*100; # substract time from TIMEOUT_MECHOKED if both clients are chocked
use constant TIMEOUT_HUNTER            => 10;     #Request a piece each X seconds                                                                {TIMEOUT_HUNTER]
use constant TIMEOUT_KEEPALIVE         => 110;    #Send a keepalive each X seconds                                                               {SEND_KEEPALIVE}
use constant TIMEOUT_KEEPALIVEDRIFT    => 80;     #Kill connection if last keepalive is older than TIMEOUT_KEEPALIVE + TIMEOUT_KEEPALIVEDRIFT    {SEND_KEEPALIVE}

## Drop the connection if we got more than X bytes in
## the clients buffer-hash (anti DOS)
use constant BUFFER_OVERFLOW => 1024*1024*5;

## Ranking:
use constant RANKING_CHOKE => -100; #When should we choke a client?
use constant RANKING_KICK  => -800; #Kick rouge clients

# RoundTrip requests:
use constant MAX_PREREQUEST_PIECES => 3;

# For people with way too many torrents:
use constant MAX_PEERS_HARDLIMIT => 400;

# Default config file
use constant DEFAULT_CONFIG_FILE => './bitflu.config';

my %BWSTATS            = (); #{UP}|{DOWN}
my %OUTQ               = (); #Outqueue
my %BFH                = (); #BufferHash (= Unpared data)
my %TCSTAT             = (); #TorrentClientStat
my %CONNECTING         = (); #BTS_CONNECTING sockets
my $TORRENT_UNIVERSE   = (); #Stores information about all torrents ; used {config} values are: seeding;uploading;downloading
my $XXX_UGLY_WOULDSEND = undef;

my %getopts = ();
GetOptions(\%getopts, "telnet=n", "http=n", "torrent=n", "config=s", "help|h", "version|v");
$getopts{config} ||= DEFAULT_CONFIG_FILE;

my $config = ();
my $rcfile = ();

if($getopts{help}) {
	die << "EOF";
Usage: $0 [--telnet 4001] [--http 4081] [--torrent 6688] [--config bitflu.config] [--help] [--version]

  -h, --help         Display this help and exit
  -v, --version      Display version information and exit
      --telnet       TCP-Port to listen for incoming telnet requests
      --http         TCP-Port to listen for incoming HTTP requests
      --torrent      TCP/UDP Port to use for BitTorrent / DHT communication
      --config       Use optional configuration file.

(C) 2006 Adrian Ulrich , report problems/bugs/patches to <adrian\@blinkenlights.ch>
Released under the terms of The "Artistic License 2.0".
Web: http://bitflu.workaround.ch

EOF
}
elsif($getopts{version}) {
	die "This is BitFlu ".BITFLU_VERSION." (Build: ".BITFLU_BUILD.")\n";
}
elsif($getopts{config}) {
	$rcfile = __load_config($getopts{config});
	$config->{client} = $rcfile;
}

#Options with getopt support:
# 1. prio: commandline switch , 2.nd: rcfile 3th: defaults
$config->{client}->{port}   = $getopts{torrent} || $rcfile->{port} ||6688;      # BitTorrent incoming port (UDP and TCP)
$config->{client}->{telnet} = $getopts{telnet}  || $rcfile->{telnet} || 4001;   # Telnet 'GUI' port
$config->{client}->{telnetbind} ||= 0;                                          # Bind telnet to interface ; defaults to 0 (= ALL)

# http can be 0 (= disabled..) so we have to do something messy:
$config->{client}->{http} = $getopts{http}  if defined($getopts{http});
$config->{client}->{http} = $rcfile->{http} if (!defined($config->{client}->{http}) && defined($rcfile->{http}));
$config->{client}->{http} = 4081 unless defined($config->{client}->{http});     # HTTP 'GUI' port
$config->{client}->{httpbind} ||= 0;                                            # Bind http-gui to interface ; defaults to 0 (= ALL)

$config->{client}->{peerid}   ||= "-BF".BITFLU_BUILD."-".sprintf("%X",int(rand(0xFFFFFFFF))).sprintf("%X",int(rand(0xFFFFFFFF)));
$config->{client}->{key}      ||= sprintf("%X",int(rand(0xFFFFFFFF))).sprintf("%X",int(rand(0xFFFFFFFF)));

 #-> cure
$config->{client}->{peerid} .= "X" x 20;
$config->{client}->{peerid} = pack("A20",$config->{client}->{peerid});
 #-> cure
$config->{client}->{key}    .= 0 x 40;
$config->{client}->{key}     = unpack("H40",$config->{client}->{key});

$config->{client}->{maxpeers}    ||= 48; #Number of connections PER TORRENT
$config->{client}->{grabpeers}   ||= 32; #Connect to maximal X peers per request
$config->{client}->{maxup}       ||= 45; #Max upload bandwidth
$config->{client}->{unchoke}     ||= 20; #how many upload slots are we providing?
$config->{client}->{rootdir}     ||= "./workdir";
$config->{client}->{autoloaddir} ||= "./autoload";
$config->{client}->{publishdir}  ||= "./publish";
$config->{client}->{kick_asktracker} ||= 60*10; #Ask trackers each X seconds
$config->{client}->{kick_autoload}   ||= 60*6;  #Scan Autoload folder each X-Seconds
#$config->{client}->{trackerblacklist} = undef;
$config->{SYS}->{boottime} = time();
$config->{SYS}->{lastcmd}  = undef;


l(L_DEBUG_CONFIG, "== CONFIGURATION ==\n".Data::Dumper::Dumper($config));


# DHT:
# should fork, secure_process() and launch DHT


##############################################################################################################################


	my $telnet_sock    = IO::Socket::INET->new(LocalPort=>$config->{client}->{telnet}, LocalAddr=>$config->{client}->{telnetbind},
	                                           Proto=>'tcp', Listen=>1, ReuseAddr=>1) 
	                                     or die "$0 Unable to bind to LocalAddr=>$config->{client}->{telnetbind}:$config->{client}->{telnet} : $!\n";
	my $telnet_handle  = new IO::Select();
	   $telnet_handle->add($telnet_sock);

	my $torrent_handle = new IO::Select();	
	my $torrent_socket  = IO::Socket::INET->new(LocalPort=>$config->{client}->{port},
	                                           Proto=>'tcp', Listen=>1, ReuseAddr=>1) or die "$0 Unable to bind to $config->{client}->{port} : $!\n";
	$torrent_handle->add($torrent_socket);


	my $http_handle = new IO::Select();
	my $http_socket = undef;
	
	if($config->{client}->{http}) {
		l(L_WARN, "** Enabling HTTP interface on $config->{client}->{httpbind}:$config->{client}->{http}");
		$http_socket = IO::Socket::INET->new(LocalPort=>$config->{client}->{http}, LocalAddr=>$config->{client}->{httpbind},
		                                           Proto=>'tcp', Listen=>1, ReuseAddr=>1) or die "$0 Unable to bind to $config->{client}->{http} : $!\n";
		$http_handle->add($http_socket);
	}
	
	
	

secure_process();

l(L_WARN,"** Init completed, now running with pid $$ ; uid = ($>|$<) / gid = ($)|$()");
l(L_WARN, "** Initializing storage-layer and contacting trackers. This may take a few seconds ...");
# Init Bitflus Store
Bitflu::Store::_INIT();
# -> Load EXISTING torrent files
__load_old_torrents();

l(L_WARN, "** Bitflu is ready! use 'telnet $config->{client}->{telnetbind} $config->{client}->{telnet}' to connect");


# -> Enter mainloop
main();


sub main {

	my ($LASTX,$NOWTIME,$KICK_ASKTRACKER, $KICK_AUTOLOAD) = undef;
	
	$KICK_ASKTRACKER = time()+$config->{client}->{kick_asktracker};
	
	# Mainloop
	for(;;) {
			
			$NOWTIME = time();
			
			
			##########################################
			# Init a non-blocking connection
			foreach my $s (keys(%CONNECTING)) {
				my $sock = $CONNECTING{$s};
				my $sin  = $TCSTAT{$sock}->{SIN};
				connect($sock, $sin);
				if($!{'EISCONN'}) {
					$torrent_handle->add($sock);
					$TCSTAT{$sock}->{BTS}     = BTS_WRITE_HANDSHAKE;
					$TCSTAT{$sock}->{TIMEOUT} = $NOWTIME+TIMEOUT_SHAKE;
					delete($CONNECTING{$sock});
				}
				else {
					if($TCSTAT{$sock}->{TIMEOUT} < $NOWTIME) {
						killSock(undef,$sock);
					}
				}
				select(undef,undef,undef,DYNASLEEP_MIN);
			}
			
			# READ torrent
			my $XXX_UGLY_READ = __bt_reader($NOWTIME);
			# WRITE torrent
			if($XXX_UGLY_WOULDSEND > 3 or ($LASTX+1 < $NOWTIME)) {
			#if(1) {
				__bt_writer($NOWTIME);
				$LASTX = $NOWTIME;
				$XXX_UGLY_WOULDSEND = undef;
			}
			
			
			if($KICK_AUTOLOAD < $NOWTIME) {
				$KICK_AUTOLOAD = $NOWTIME+$config->{client}->{kick_autoload};
				__auto_loader(Bitflu::Store::__get_autoloadbase());
			}
			
			
			# Fixme: .. bler hack: Wir sollten TORRENT_UNIVERSE checken und PERIODIC-Dingsens setzen
			if($KICK_ASKTRACKER < $NOWTIME) {
				$KICK_ASKTRACKER = $NOWTIME+$config->{client}->{kick_asktracker};
				foreach my $xt (shuffle(keys(%$TORRENT_UNIVERSE))) {
					print "? -> $xt\n";
					Bitflu::RPC::execute("qtracker $xt");
					last;
				}
				kill(1,$$); #..and save config
			}
			
			## TELNET START
			__telnet_reader($NOWTIME);
			__telnet_writer($NOWTIME);
			## TELNET END
			
			__http_reader($NOWTIME);
			__http_writer($NOWTIME);
			my $XXX_UGLY_SLEEP = DYNASLEEP_MAX;
			if($XXX_UGLY_READ >= POSIX::BUFSIZ) {
				$XXX_UGLY_SLEEP = DYNASLEEP_MIN;
			}
			___xxx_ugly_genbwstats($NOWTIME);
			select(undef,undef,undef,$XXX_UGLY_SLEEP);
	} #END mainloop
	
}

########################################################################
# Chroots and drops privileges
sub secure_process {
	if($config->{client}->{chroot}) {
		l(L_WARN, "+ Chrooting into $config->{client}->{chroot}");
		Carp::cluck("CHROOT_CARP_INIT");
		chdir($config->{client}->{chroot})  or Carp::confess("chdir($config->{client}->{chroot}) failed : $!");
		chroot($config->{client}->{chroot}) or Carp::confess("chroot($config->{client}->{chroot}) failed : $!");
	}
	else {
		l(L_WARN, "- No chroot option set :-( ; starting up without jail");
	}
	
	if($config->{client}->{runas_gid}) {
		l(L_WARN, "+ Changing GID to $config->{client}->{runas_gid}");
		$! = undef;
		$) = "$config->{client}->{runas_gid} $config->{client}->{runas_gid}";
			Carp::confess("Unable to set EGID: $!") if $!;
		
		$! = undef;
		$( = "$config->{client}->{runas_gid}";
			Carp::confess("Unable to set GID: $!") if $!;
	}
	
	if($config->{client}->{runas_uid}) {
		l(L_WARN, "+ Chaging UID to $config->{client}->{runas_uid}");
		POSIX::setuid($config->{client}->{runas_uid}) or Carp::confess("Unable to set UID : $!");
	}
	
	Carp::confess("$0 refuses to run with uid 0 (= root)\n") unless $>;
}


########################################################################
# Load existing .torrent files and init $TORRENT_UNIVERSE
sub __load_old_torrents {
	opendir(OLD_TORRENTS, Bitflu::Store::__get_torrentbase()) or Carp::confess("Unable to open torrentsbase : No init?");
	while(my $file = readdir(OLD_TORRENTS)) {
		my $fp = Bitflu::Store::__get_torrentbase()."/$file";
		next unless (-f $fp);
		my $info_hash = __load_torrent($fp);
		
		unless($info_hash) {
			l(L_WARN, "** BROKEN torrent file in workdir: $fp");
		}
	}
	closedir(OLD_TORRENT);
}

########################################################################
# Load a single .torrent file from disk
sub __load_torrent {
	my($file) = @_;
	my $hash = Bitflu::Bencoding::torrent2hash($file);
	if($hash->{torrent_hash} =~ /^([a-z0-9]{40})$/ && !defined($TORRENT_UNIVERSE->{$hash->{torrent_hash}})) {
		$TORRENT_UNIVERSE->{$hash->{torrent_hash}}->{torrent} = \$hash;
		$TORRENT_UNIVERSE->{$hash->{torrent_hash}}->{sourcefile} = Bitflu::Store::StoreTorrent($file, $hash->{torrent_hash});
		Bitflu::Store::InitTorrent($hash->{torrent_hash});
		#Overwrite zero-conf that InitTorrent setup
		$TORRENT_UNIVERSE->{$hash->{torrent_hash}}->{config}     = Bitflu::Store::GetConfig($hash->{torrent_hash});
		Bitflu::Tracker::Query($hash->{torrent_hash});
		BT_getPeers($hash->{torrent_hash});
		return $hash->{torrent_hash};
	}
	else {
		return undef;
	}
}

########################################################################
# Parse configuration file
sub __load_config {
	my($file) = @_;
	my %h = ();
	
	unless(open(CONF,$file)) {
		warn "$0 : Unable to open configuration file ($file) ; Continuing with default options\n";
		return \%h;
	}
	while(<CONF>) {
		my $line = $_;
		chomp($line);
		
		if($line =~ /^#/ or $line =~ /^\s*$/) {
			#comment
			next;
		}
		elsif($line =~ /^([a-zA-Z_]+)\s*=\s*(.+)$/) {
			die "$0 Duplicate definition of '$1' at line $. in configuration file '$file'\n" if $h{$1};
			$h{$1} = $2;
		}
		else {
			die "$0: Mumbled line '$line' at line $. in configuration file '$file'\n";
		}
	}
	close(CONF);
	return \%h;
}


########################################################################
# Create Bandwith Statistics: Fixme: This needs a redesign ;-)
sub ___xxx_ugly_genbwstats {
	my($NOWTIME) = @_;
			if((my $drift = ($NOWTIME - $BWSTATS{LASTCHECK}))) {
				$BWSTATS{__UPX}   = $BWSTATS{_UPX};
				$BWSTATS{__DOWNX} =$BWSTATS{_DOWNX};
				$BWSTATS{_UPX} = $BWSTATS{UPX};
				$BWSTATS{_DOWNX} = $BWSTATS{DOWNX};
				$BWSTATS{UPX} = (($BWSTATS{UP}/$drift/1024)+$BWSTATS{_UPX}+$BWSTATS{__UPX})/3;
				$BWSTATS{DOWNX}=(($BWSTATS{DOWN}/$drift/1024)+$BWSTATS{_DOWNX}+$BWSTATS{_DOWNX})/3;
				$BWSTATS{UP} = 0;
				$BWSTATS{DOWN}=0;
				$BWSTATS{RUN} = 0;
				$BWSTATS{LASTCHECK}=$NOWTIME
			}
}


########################################################################
# Check how many connections this torrent has
sub CheckSingleNodeConnectX {
	my($hash,$ip) = @_;
	Carp::confess("INVALID TORRENT HASH") unless defined($TORRENT_UNIVERSE->{$hash});
	my $xnum = 0;
	   $xnum = keys(%{$TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip}})
	           if defined($TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip});
	return $xnum;
}

########################################################################
# Add a connection for this ip to torrent
sub AddSingleNodeConnectX {
	my($hash,$ip,$glob) = @_;
	Carp::confess("INVALID TORRENT HASH") unless defined($TORRENT_UNIVERSE->{$hash});
	Carp::confess("Doubleadd?") if defined($TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip}->{$glob});
	Carp::confess("No IP?")     unless $ip;
	Carp::confess("No Sock?")   unless $glob;
	return ($TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip}->{$glob}++);
}



########################################################################
# Save conf
sub ConfDown {
	my($sig) = @_;
	l(L_WARN, "Got signal $sig : saving configuration");
	# Save config before going down
	foreach my $k (keys(%$TORRENT_UNIVERSE)) {
		l(L_WARN, " => Saving Configuration of '$k'");
		Bitflu::Store::SaveConfig($k, $TORRENT_UNIVERSE->{$k}->{config});
	}
	
	if($sig eq 'INT') {
		l(L_WARN, "Bitflu shutdown completed");
		exit(0);
	}
}


########################################################################
# Kill connection with a client and cleanup all buffers
sub killSock {
	my($select_handle,$socket) = @_;
	l(L_DEBUG, "killSock($select_handle, $socket)");
	$select_handle->remove($socket) if $select_handle;
	delete $BFH{$socket};
	delete $OUTQ{$socket};
	FreeWorkPieces($TCSTAT{$socket});
	
	if($TCSTAT{$socket}->{INFO_HASH}) {
		Carp::confess("Invalid INFO_HASH") unless $TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}};
		Carp::confess("No IP?!")           unless $TCSTAT{$socket}->{IP};
		
		if ( defined($TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}}->{peerslist}->{$TCSTAT{$socket}->{IP}}) ) {
			# IP in Storage
			delete($TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}}->{peerslist}->{$TCSTAT{$socket}->{IP}}->{$socket});
			if(CheckSingleNodeConnectX($TCSTAT{$socket}->{INFO_HASH}, $TCSTAT{$socket}->{IP}) == 0) {
				delete($TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}}->{peerslist}->{$TCSTAT{$socket}->{IP}}); #Freeup some memory
			}
		}
		
	}
	
	delete $TCSTAT{$socket};
	delete $CONNECTING{$socket};
	close $socket or Carp::confess("DIEDIEDIE <$socket> : $! => $?");
}



########################################################################
# Get data read from client
sub _xGETtoread {
	my($sock) = @_;
	return $BFH{$sock};
}

########################################################################
# Clean (delete) data read from client
sub _xCLEANtoread {
	my($sock,$to) = @_;
	if($to < 0) { delete $BFH{$sock} ; return }
	$BFH{$sock} =  substr($BFH{$sock},$to);
}


########################################################################
# STORE readed data APPEND
sub _xrA {
	my($sock,$buff) = @_;
	$BFH{$sock} .= $buff;
}

########################################################################
# STORE Readed data NEW
sub _xrN {
	my($sock,$buff) = @_;
	$BFH{$sock} = $buff;
}

###

########################################################################
# Clean (delete) data read from client
sub _xCLEANtowrite {
	my($sock,$to) = @_;
	if($to < 0) { delete $OUTQ{$sock}; return }
	$OUTQ{$sock} =  substr($OUTQ{$sock},$to);
}

########################################################################
# Get data needed to send to client
sub _xGETtowrite {
	my($sock) = @_;
	return $OUTQ{$sock};
}

########################################################################
# WRITE data NEW
sub _xwN {
	my($sock, $txt) = @_;
	$XXX_UGLY_WOULDSEND++;
	$OUTQ{$sock} = $txt;
}

########################################################################
# WRITE data Append
sub _xwA {
	my($sock, $txt) = @_;
	$XXX_UGLY_WOULDSEND++;
	$OUTQ{$sock} .= $txt;
}

########################################################################
# Print a logmsg
sub l {
	my($level, @message) = @_;
	printf("".gmtime()." %03d # %s\n", $level, "@message") if $level >= LOGLEVEL;
	die if $level == L_PANIC;
}


########################################################################
# Connect to a BT-Client and populate TCSTAT for this connection
sub BT_initConnection {
	my($hr) = @_;
	
	Carp::confess("Called without {hash}") unless $hr->{hash};
	Carp::confess("Invalid hash") unless $TORRENT_UNIVERSE->{$hr->{hash}};
	
	if(CheckSingleNodeConnectX($hr->{hash}, $hr->{ip})) {
		l(L_DEBUG, "Outgoing connection for $hr->{ip} ==> $hr->{hash} not established: We allready got one!");
		return undef;
	}
	
	my $proto = getprotobyname('tcp');
	my $sock  = undef;
	socket($sock, AF_INET,SOCK_STREAM,$proto) or Carp::confess("Unhandled state : $!\n");
	my $sin = sockaddr_in($hr->{port}, inet_aton($hr->{ip}));
	unblock_FH($sock) or Carp::confess("Unable to unblock FH <$sock> : $!\n");

	$CONNECTING{$sock} = $sock;
	
	$TCSTAT{$sock} and Carp::confess("BUGBUG: <$sock> was closed but somehow \$TCSTAT was still alive! PANICED");
	$TCSTAT{$sock}->{BTS} = BTS_CONNECTING;
	$TCSTAT{$sock}->{SIN} = $sin;
	$TCSTAT{$sock}->{IP}  = $hr->{ip};
	$TCSTAT{$sock}->{INFO_HASH}       = $hr->{hash};
	$TCSTAT{$sock}->{ME_CHOKE}        = 1;
	$TCSTAT{$sock}->{PEER_CHOKE}      = 1;
	$TCSTAT{$sock}->{ME_INTERESTED}   = 0;
	$TCSTAT{$sock}->{PEER_INTERESTED} = 0;
	$TCSTAT{$sock}->{TIMEOUT}         = time() + TIMEOUT_SHAKE;
	$TCSTAT{$sock}->{SEND_KEEPALIVE}  = time() + TIMEOUT_KEEPALIVE;
	AddSingleNodeConnectX($hr->{hash},$hr->{ip},$sock);
}


########################################################################
# Connect to some peers
sub BT_getPeers{
	my($hash) = @_;
	my $torrent = $TORRENT_UNIVERSE->{$hash} or Carp::confess("Called with uninitialized torrent?!");
	my $peerslots_left = GetMaxConnections() - int(keys(%TCSTAT));
	   $peerslots_left = $config->{client}->{grabpeers} if $peerslots_left > $config->{client}->{grabpeers};
	
	l(L_WARN, "BT_getPeers($hash) : Slots set to $peerslots_left");
	
	if(ref($torrent->{announce}) ne "HASH") {
		l(L_DEBUG, "BT_getPeers($hash) : Empty {announce} ; tracker not asked / dead / DHTed");
		return undef;
	}
	
	foreach my $r (shuffle(@{$torrent->{announce}->{peers}})) {
		last if $peerslots_left-- < 1;
		BT_initConnection({port=>$r->{port}, ip=>$r->{ip}, hash=>$hash});
		l(L_DEBUG, "BT_getPeers($hash) : $peerslots_left => $r->{ip}");
	}
}

########################################################################
# Returns (dynamic) number of max-connections
sub GetMaxConnections {
	my $num_torrents = int(keys(%$TORRENT_UNIVERSE));                  # Number of torrents we are downloading
	my $num_maxconn  = $num_torrents * $config->{client}->{maxpeers};  # Maximal connections possible
	   $num_maxconn  = MAX_PEERS_HARDLIMIT if $num_maxconn > MAX_PEERS_HARDLIMIT;
	return $num_maxconn;
}


########################################################################
# Unblock a filehandle
sub unblock_FH {
	my($cfh) = @_;
	return undef unless $cfh;
	my $flags = fcntl($cfh, F_GETFL, 0)
							or return undef;
	fcntl($cfh, F_SETFL, $flags | O_NONBLOCK)
	or return undef;
	return 1;
}


########################################################################
# Scan given folder for files and try to load them all as .torrent
sub __auto_loader {
	my($xdir) = @_;
	opendir(AUTOLOAD, $xdir) or Carp::confess("Unable to open AutoLoad folder $xdir : $!");
	while(my $toload = readdir(AUTOLOAD)) {
		my $fullpath = $xdir."/".$toload;
		next unless -f $fullpath;
		my $result = __load_torrent($fullpath);
		if($result) {
			l(L_WARN, "[AutoLoad] '$fullpath' loaded (Hash: $result)");
			unlink($fullpath) or Carp::cluck("Unable to remove '$fullpath' : $!");
		}
		else {
			l(L_WARN, "[AutoLoad] '$fullpath' NOT loaded due to error");
		}
	}
	closedir(AUTOLOAD);
}


sub __http_reader {
	my @all_http_can_read = $http_handle->can_read(0);
	foreach my $sock (@all_http_can_read) {
		if($sock == $http_socket) {
			#-> A new client
			my $new_sock = $sock->accept();
			$http_handle->add($new_sock);
		}
		else {
			#-> Command from connected client
			my $buffer = undef;
			my $recv = $sock->recv($buffer, POSIX::BUFSIZ,0);
			my $bufflen = length($buffer);
			if(defined($recv) and $bufflen > 0) {
				_xrA($sock,$buffer); #Append buffer
				my $full_data = _xGETtoread($sock);
				if(substr($full_data,-4,4) eq "\r\n\r\n") {
					_xCLEANtoread($sock,-1);
					_xwA($sock, Bitflu::HTTPgui::main($full_data));
				}
			}
			else {
				killSock($http_handle,$sock);
			}
		}
	}
}

########################################################################
# Write output to telnet client
sub __http_writer {
	my @all_http_can_write = $http_handle->can_write(DYNASLEEP_MIN);
	my $clients = 0;
	foreach my $sock (@all_http_can_write) {
		$clients++;
		if(my $buff = _xGETtowrite($sock)) {
			my $data_sent = syswrite($sock,$buff);
			_xCLEANtowrite($sock,$data_sent);
			if(!defined($data_sent)) {
				l(L_DEBUG, "Syswrite to <$sock> failed: $! ; Closing connection with peer");
				killSock($http_handle,$sock);
			}
			elsif($data_sent == length($buff)) { # Buffer empty, close connection
				killSock($http_handle,$sock);
			}
		}
	}
	return $clients;
}


########################################################################
# Read telnet commands
sub __telnet_reader {
			my @all_telnet_can_read = $telnet_handle->can_read(0);
			foreach my $sock (@all_telnet_can_read) {
				if($sock == $telnet_sock) {
					#-> A new client
					my $new_sock = $sock->accept();
					$telnet_handle->add($new_sock);
					_xwN($new_sock, "## Welcome to BitFlu\r\n# ");
					l(L_WARN,"[Telnet] New client <$sock> connected (".$new_sock->peerhost().")");
				}
				else {
					#-> Command from connected client
					my $buffer = undef;
					my $recv = $sock->recv($buffer, POSIX::BUFSIZ,0);
					my $bufflen = length($buffer);
					if(defined($recv) and $bufflen > 0) {
						_xrA($sock,$buffer); #Append buffer
						my $full_data = _xGETtoread($sock);
						_xCLEANtoread($sock,-1);
						if(substr($full_data,-2,2) eq "\r\n") {
							Bitflu::Telnet::parseCMD($telnet_handle,$sock,$full_data);
							_xwA($sock, "\r# ");
						}
					}
					else {
						l(L_INFO, "[Telnet] Closing connection with <$sock>");
						killSock($telnet_handle,$sock);
					}
				}
			}
}


########################################################################
# Write output to telnet client
sub __telnet_writer {
	my @all_telnet_can_write =$telnet_handle->can_write(DYNASLEEP_MIN);
	my $clients = 0;
	foreach my $sock (@all_telnet_can_write) {
		$clients++;
		if(my $buff = _xGETtowrite($sock)) {
			my $data_sent = syswrite($sock,$buff);
			_xCLEANtowrite($sock,$data_sent);
			if(!defined($data_sent)) {
				l(L_DEBUG, "Syswrite to <$sock> failed: $! ; Closing connection with peer");
				killSock($telnet_handle,$sock);
			}
		}
	}
	return $clients;
}


########################################################################
# Read data from BT clients
sub __bt_reader {
	my($NOW) = @_;
			my @all_bt_can_read = $torrent_handle->can_read(0);
			my $XXX_UGLY_ALL_LEN = 0;
			foreach my $sock (@all_bt_can_read) {
				if($sock == $torrent_socket) {
				# -> New client
					my ($new_sock) = $sock->accept();
					my  $new_ip = undef;
					
					if(!defined($new_sock) or !($new_ip = $new_sock->peerhost())) {
						l(L_DEBUG, "$sock -> accept() failed ; removing socket");
						$new_sock->close() if $new_sock; # Let IO::Socket take care of this mess
						next;
					}
					elsif(!unblock_FH($new_sock)) { #For BSD
						Carp::cluck("Unable to unblock FH <$new_sock>");
						$new_sock->close() if $new_sock;
						next;
					}
					
					#Connection established: Build a small ENV for this client
					$torrent_handle->add($new_sock);
					$TCSTAT{$new_sock} and Carp::confess("BUGBUG: This is a duplicate!"); #Paranoia check
					$TCSTAT{$new_sock}->{BTS} = BTS_KILLME_PLEASE;
					
					#Check if we have a slot for the client
					if(int(keys(%TCSTAT)) >= GetMaxConnections() ) {
						l(L_WARN, "maxpeers reached; Closing connection with new client: $new_ip");
						# -> BTS is set to BTS_KILLME_PLEASE : this will kill the client in the next run of bt_writer
					}
					else {
					l(L_INFO, "<$new_sock> Torrent client connected, waiting for HANDSHAKE");
						# Fixme: BT_Init has almost the same code.. 
						$TCSTAT{$new_sock}->{BTS}             = BTS_READ_CLIENT_HANDSHAKE;
						$TCSTAT{$new_sock}->{IP}              = $new_ip;
						$TCSTAT{$new_sock}->{ME_CHOKE}        = 1;
						$TCSTAT{$new_sock}->{PEER_CHOKE}      = 1;
						$TCSTAT{$new_sock}->{ME_INTERESTED}   = 0;
						$TCSTAT{$new_sock}->{PEER_INTERESTED} = 0;
						$TCSTAT{$new_sock}->{TIMEOUT}         = $NOW+TIMEOUT_SHAKE;
						$TCSTAT{$new_sock}->{SEND_KEEPALIVE}  = $NOW+TIMEOUT_KEEPALIVE;
					}
				}
				else {
				#-> Data received from client
					my $buffer        = undef;
					my $recv          = recv($sock,$buffer, POSIX::BUFSIZ,0);
					my $bufflen       = length($buffer);
					my $tref          = $TCSTAT{$sock};
					if(defined($recv) and $bufflen > 0) {
						_xrA($sock,$buffer); #Append buffer
						
						RELOOP:
						my $full_data     = _xGETtoread($sock);
						my $full_data_len = length($full_data);
						my $state         = $tref->{BTS};
						$XXX_UGLY_ALL_LEN += $full_data_len;
						
						$tref->{RECEIVED_KEEPALIVE} = $NOW + TIMEOUT_KEEPALIVE; # Message received!
						
						if($state == BTS_IDLE) {
							if($full_data_len >= 4) { #Minimal length of a BT message
								my $msglen = unpack("N", substr($full_data,0,4));
								
								if($msglen == 0) {
									l(L_DEBUG,"   <- KEEPALIVE received");
									# No need to update RECEIVED_KEEPALIVE as we did this a few lines ago :-)
									_xCLEANtoread($sock,4); #Remove from buffer
									goto RELOOP;  #Pickup more
								}
								elsif(length($full_data) >= ($msglen+4)) {
									# -> We got a full packet with payload!
									my $msgtype = unpack("c", substr($full_data,4,1));
									_xCLEANtoread($sock, $msglen+4); #Remove parsed data from buffer
									
									###################################################################
									if($msgtype == 7) { # Piece
										my $piece_id          = unpack("N", substr($full_data,5,4));
										my $offset            = unpack("N", substr($full_data,9,4));
										my $content           = substr($full_data,13,$msglen-9);
										my $current_info_hash = $tref->{INFO_HASH};
										
										# .. should NEVEREVER happen:
										Carp::confess("BUGBUG: Invalid piecelen") if length($content)+9  != $msglen;
										
										# Add some statistic:
										$BWSTATS{DOWN} += $msglen;
										$TORRENT_UNIVERSE->{$current_info_hash}->{config}->{downloaded} += $msglen-9;
										
										# ..and increment the clients timeout
										$tref->{TIMEOUT_WAITFORPIECE} = $NOW + TIMEOUT_WAITFORPIECE;
										
										
										if(!defined($tref->{PIECE_INWORK}{$piece_id})) {
											# -> This client sent us something we didn't request.
											# We are going to DROP this.
											# Storing it anyway wouldn't be a good option because this client has to be considered
											# 'rouge' => The data sent may be even faulty (SHA1 corruption) : Such a case would force
											# us to download the whole piece again!
											l(L_WARN, "<$sock> sent us piece an unrequested piece: $piece_id\@$current_info_hash ; Killing connection with 'rouge' client.");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
										
										
										# Attemp to store data
										my $stored = Bitflu::Store::StorePiece({hash=>$current_info_hash, index=>$piece_id, offset=>$offset, data=>$content});
										
										if(defined($stored->{verified})) {
											# Chunk completed + Verified
											l(L_DEBUG_STORE, " => Piece $piece_id\@$current_info_hash is VERIFIED (got it from $sock)");
											Bitflu::Bitfield::SetBit(\$TORRENT_UNIVERSE->{$current_info_hash}->{bitfield},$piece_id);
											
											# Pollute the network with funny HAVE messages
											foreach my $sname (keys(%TCSTAT)) {
												my $ref = $TCSTAT{$sname};
												if($ref->{BTS} == BTS_IDLE && $ref->{INFO_HASH} eq $current_info_hash &&
												   Bitflu::Bitfield::GetBit(\$ref->{BITFIELD},$piece_id) == 0) {
													l(L_DEBUG, "  --> Sending *HAVE $piece_id* to $sname");
													_xwA($sname, Bitflu::Protocol::CreateHave($piece_id));
												}
											}
										}
										
										if($stored->{rollover}) {
											# -> Storemanager told us to do a RollOver (= Grab next piece, not the next offset)
											
											delete($tref->{PIECE_INWORK}{$piece_id}); # Storemanager saved piece, so delete thisone..
											my $suggested_piece = Bitflu::Store::SuggestNewPiece($tref,$piece_id); #Suggest a new piece but NOT $piece_id
											l(L_INFO, "StoreManager suggested to try a new piece -> $suggested_piece\@$tref->{INFO_HASH} (old: $piece_id)");
											if(defined($suggested_piece)) {
												#Change store
												$stored->{index} = $suggested_piece;
												($stored->{offset}, $stored->{size} ) = Bitflu::Store::GetFreePieceOffset($tref->{INFO_HASH},$stored->{index});
												Bitflu::Store::SetPieceAsInwork($tref->{INFO_HASH},$stored->{index});
												$tref->{PIECE_INWORK}{$stored->{index}} = $stored->{index};
											}
											elsif(int(keys(%{$tref->{PIECE_INWORK}})) > 0) {
												l(L_WARN, "Client has NO new pieces for me but some requests are still unfullfilled ; ");
												goto RELOOP;
											}
											else {
												l(L_WARN, "Client has NO pieces left for $current_info_hash ");
												$tref->{ME_CHOKE} = 1;
												$tref->{TIMEOUT_MECHOKED} = $NOW + TIMEOUT_MECHOKED;
												_xwA($sock, Bitflu::Protocol::CreateNotInterested());
												FreeWorkPieces($tref);
												goto RELOOP; #Fixme: This is ugly but we need to aviod sending a request!
											}
										}
										
										# Offset problem may be here for the last piece if Store fucked up..
										
										
										_xwA($sock, Bitflu::Protocol::CreateRequest($stored));
										$tref->{PIECE_REQUEST} = $stored->{index};
										$tref->{RANKING} += $stored->{points}+1;
									}
									elsif($msgtype == 4) {
									# -> HAVE
										my $piece_id = unpack("N", substr($full_data,5,4));
										l(L_DEBUG, "<$sock> set HAVE $piece_id");
										Bitflu::Bitfield::SetBit(\$tref->{BITFIELD},$piece_id);
										
										unless(Bitflu::Bitfield::GetBit(\$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{bitfield},$piece_id)) {
											#Trigger a search
											$tref->{TIMEOUT_INTERESTED} = 0;
										}
										
									}
									elsif($msgtype == 5) {
									# -> Bitfield
										my $bitfield = substr($full_data,5,$msglen-1);
										$tref->{BITFIELD} = $bitfield;
										
										my $len_bf = length(unpack("B*",$bitfield));                          # Worst case: 1
										my $parts  = Bitflu::Store::GetNumberOfPieces($tref->{INFO_HASH});    # Worst case: 0
										if($len_bf < ($parts+1)) {
											l(L_WARN, "<$sock> Client sent a faulty bitfield ; Closing connection");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
									}
									elsif($msgtype == 1) {
									# -> Client unchoked me!
										if($tref->{ME_CHOKE}) {
											$tref->{ME_CHOKE} = 0;
											l(L_DEBUG, "<$sock> Got unchoked :-)");
										}
									}
									elsif($msgtype == 0) { ## -> Choke
										l(L_DEBUG, "<$sock> I am choked :-(");
										if($tref->{ME_CHOKE} == 0) {
										# -> Switch
											$tref->{ME_CHOKE} = 1;
											$tref->{TIMEOUT_MECHOKED} = $NOW + TIMEOUT_MECHOKED;
											# Do we have a queued piece? If yes -> Drop'em
											FreeWorkPieces($tref);
										}
									}
									elsif($msgtype == 2) { ## -> interested
										l(L_DEBUG, "<$sock> Unchoking client?");
										
										#Fixme: This doesn't scale well : GetNumberOfUnchokedPeers is expensive!
										my $open_slots = $config->{client}->{unchoke} - GetNumberOfUnchokedPeers();
										
										if($open_slots > 0) {
											$tref->{PEER_CHOKE} = 0;
											$tref->{RANKING} += 5;
											l(L_DEBUG, "<$sock> .. yes -> $open_slots");
											_xwA($sock, Bitflu::Protocol::CreateUnchoke());
										}
										else {
											l(L_DEBUG, "<$sock> .. nope -> $open_slots");
										}
									}
									elsif($msgtype == 3) { ## -> Not interested
										l(L_DEBUG, "<$sock> Client choked itself");
										$tref->{PEER_CHOKE} = 1;
									}
									elsif($msgtype == 6) { ## -> request
										my $piece_id = unpack("N", substr($full_data,5,4));
										my $offset   = unpack("N", substr($full_data,9,4));
										my $length   = unpack("N", substr($full_data,13,4));
										l(L_DEBUG, "<$sock> sent a request for piece $piece_id at offset $offset with len $length");
										if($length > MAXPIECESIZE) {
											l(L_WARN, "<$sock> Requested too much data for $piece_id -> $length . Closing connection");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
										elsif($tref->{PEER_CHOKE} == 0) {
											my $ref = {offset=>$offset, index=>$piece_id, size=>$length, hash=>$tref->{INFO_HASH}};
											l(L_DEBUG, "<$sock> GET $ref->{index}\@$tref->{INFO_HASH}");
											$tref->{STATS_UPLOADED} += $length;
											$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{config}->{uploaded} += $length;
											my ($piece_data, $piece_len) = Bitflu::Store::ReadPiece($ref);
											if($piece_len != $ref->{size}) {
												l(L_WARN, "<$sock> Requested piece $piece_id at offset $offset with a length of $length. Reading data failed after $piece_len bytes");
												#drop client
												$tref->{BTS} = BTS_KILLME_PLEASE;
												next;
											}
											else {
												_xwA($sock, Bitflu::Protocol::CreatePiece($ref).$piece_data);
												# Penalty downloading client UNLESS we are seeding this torrent.
												$tref->{RANKING}-- unless $TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{config}->{seeding};
											}
										}
										else {
											l(L_WARN, "<$sock> Choked client asked for $piece_id !");
											$tref->{RANKING} -= 10;
										}
									}
									elsif($msgtype == 8) {
										# void(); Cancel isn't supported .. sorry: i think the endgame stuff is really stupid.
									}
									elsif($msgtype == 9) { #DHT PORT message
										my $dht_port = unpack("n", substr($full_data,5,2));
										l(L_WARN, "[DHT] New bootnode for $tref->{INFO_HASH} : ".$tref->{IP}.":$dht_port");
										Bitflu::Store::AddDhtNode($tref->{INFO_HASH}, $tref->{IP}, $dht_port);
									}
									else {
										l(L_WARN, "  <-> Unknown msgtype $msgtype ".unpack("H*",$msgtype));
									}
									
									goto RELOOP;
								}
								else {
									# -> Not a complete chunk.. wait..
								}
								
							}
							
						}
						elsif($state == BTS_READ_HANDSHAKE_RESPONSE) {
							my $hsresp = Bitflu::Protocol::ParseHandshake($full_data);
							_xrN($sock,$hsresp->{ol}); #Save unparsed data
							
							if($hsresp->{info_hash} eq $tref->{INFO_HASH}) {
								l(L_INFO, "<$sock> Handshake finished, info hash -> $tref->{INFO_HASH}");
								$tref->{CLIENT_ID} = $hsresp->{client_id};
								$tref->{BTS}       = BTS_WRITE_BITFIELD;
								$tref->{TIMEOUT}   = $NOW+TIMEOUT_SHAKE;
								#We may have unparsed data:
								goto RELOOP;
							}
						}
						elsif($state == BTS_READ_CLIENT_HANDSHAKE) {
							my $hsresp = Bitflu::Protocol::ParseHandshake($full_data);
							_xrN($sock,$hsresp->{ol}); #Save unparsed data
							if($TORRENT_UNIVERSE->{$hsresp->{info_hash}}) { # Shared file!
								if($hsresp->{client_id} eq $config->{client}->{peerid}) {
									#Whoot: This client (connected itself to us) has the same peerid as we got..
									# Most likely we are connecting to ourself:
									l(L_INFO, "<$sock> sent my own peerid ($config->{client}->{peerid}) ; Looks like i've connected to myself! dropping connection!");
									$tref->{BTS} = BTS_KILLME_PLEASE;
									next;
								}
								elsif(CheckSingleNodeConnectX($hsresp->{info_hash}, $tref->{IP})) {
									l(L_DEBUG, "<$sock> $tref->{IP} has a connection with $hsresp->{info_hash} ; closing 2nd connection");
									$tref->{BTS} = BTS_KILLME_PLEASE;
									next;
								}
								else {
									l(L_INFO, "<$sock> Handshake parsed, sending my own bitfield to client");
									$tref->{CLIENT_ID}  = $hsresp->{client_id};
									# -> Connection not initiated by us? -> Write response and bitfield
									$tref->{BTS}        = BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD;
									$tref->{TIMEOUT}    = $NOW+TIMEOUT_SHAKE;
									$tref->{INFO_HASH}  = $hsresp->{info_hash};
									AddSingleNodeConnectX($hsresp->{info_hash}, $tref->{IP}, $sock);
									goto RELOOP;
								}
							}
						}
						else {
							l(L_DEBUG,"Unhandled state: $state");
						}
					}
					else {
						l(L_INFO, "bt : closing connection with <$sock>");
						#Note: we can't use BTS_KILLME_PLEASE here as the socket
						# IS allready dead..
						killSock($torrent_handle,$sock);
					}
				}
			}
	return $XXX_UGLY_ALL_LEN;
}

########################################################################
# Write data to BT clients
sub __bt_writer {
	my($NOW) = @_;
	my @all_bt_can_write =$torrent_handle->can_write(DYNASLEEP_MIN);
	
	___xxx_ugly_genbwstats($NOW);
	my $xleft = ($config->{client}->{maxup}-$BWSTATS{UPX});
	   $xleft = 1 if $xleft < 1;
	my $upcredits = 4+int($xleft*1024/(1+GetNumberOfUnchokedPeers()));
	
	my $clients = 0;
	foreach my $sock (@all_bt_can_write) {
		my $tref = $TCSTAT{$sock};
		my $state = $tref->{BTS};
		$clients++;
		
		
		if($state == BTS_IDLE) {
			#We are mostly here :-)
			#########################################################
			# -> Check if we need to send a keepalive

			### => Keepalive checks
			if($tref->{RECEIVED_KEEPALIVE} <= ($NOW - TIMEOUT_KEEPALIVEDRIFT)) {
				l(L_DEBUG, "<$sock> Timed out: No keepalive!");
				killSock($torrent_handle,$sock);
				next;
			}
			elsif($tref->{ME_CHOKE} && $tref->{PEER_CHOKE} &&
			          ($tref->{TIMEOUT_MECHOKED} - TIMEOUT_DOUBLECHOKE_DRIFT) < $NOW) {
				l(L_DEBUG, "<$sock> -> Closing DOUBLECHOKE state");
				killSock($torrent_handle,$sock);
				Carp::confess("DOUBLECHOKEASSERTFAULTED") if TIMEOUT_DOUBLECHOKE_DRIFT >= TIMEOUT_MECHOKED;
				next;
			}
			elsif($tref->{ME_CHOKE} && $tref->{TIMEOUT_MECHOKED} < $NOW && !$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{config}->{seeding}) {
				l(L_WARN, "<$sock> -> Closing connection because i am still choked :-(");
				killSock($torrent_handle,$sock);
				next;
			}
			elsif(int(keys(%{$tref->{PIECE_INWORK}})) > 0 && $tref->{TIMEOUT_WAITFORPIECE} < $NOW) {
				l(L_WARN, "<$sock> -> Closing connection : Waited much too long for piece $tref->{PIECE_INWORK}\@$tref->{INFO_HASH}");
				killSock($torrent_handle,$sock);
				next;
			}
			elsif($tref->{SENT_KEEPALIVE} <= $NOW) {
				l(L_DEBUG, "<$sock> -> Sending a Keepalive");
				$tref->{SENT_KEEPALIVE} = $NOW + TIMEOUT_KEEPALIVE;
				_xwA($sock, Bitflu::Protocol::CreateKeepalive());
			}
			elsif($tref->{RANKING} < RANKING_KICK) {
				l(L_WARN, "<$sock> Kicking client with ranking $tref->{RANKING}");
				killSock($torrent_handle,$sock);
				next;
			}
			elsif($tref->{PEER_CHOKE} == 0 && $tref->{RANKING} < RANKING_CHOKE) {
				l(L_DEBUG, "<$sock> -> Choking client with ranking $tref->{RANKING}");
				$tref->{PEER_CHOKE} = 1; #= Stop sending me requests
				_xwA($sock, Bitflu::Protocol::CreateChoke());
			}
			
			
			## tick tack ##
			
			if($tref->{ME_CHOKE} && $tref->{TIMEOUT_INTERESTED} < $NOW) {
				# -> I am choked
				l(L_DEBUG_SUGST, "<$sock> Checking if client has data for me");
				$tref->{TIMEOUT_INTERESTED} = $NOW + TIMEOUT_INTERESTED; #Set timer for next check
				
				if(defined(Bitflu::Store::SuggestNewPiece($tref,undef))) {
					# -> And client would have data for us :-) -> Send an interested request to him
					_xwA($sock, Bitflu::Protocol::CreateInterested());
					l(L_DEBUG, "<$sock> Yes, sent an INTERESTED message to peer");
				}
				else {
					#Bad dog: Wait some time until we try again:
					$tref->{TIMEOUT_INTERESTED} += int(rand(TIMEOUT_INTERESTED));
					
					#Last resort: go not too high..
					if($tref->{TIMEOUT_INTERESTED} >= $NOW+TIMEOUT_INTERESTED_MAX) {
						$tref->{TIMEOUT_INTERESTED} = $NOW+int(rand(TIMEOUT_INTERESTED_MAX));
					}
					l(L_DEBUG_SUGST,"<$sock> got a PENALTY: ".gmtime($tref->{TIMEOUT_INTERESTED})."");
				}
			}
			elsif($tref->{ME_CHOKE} == 0 && $tref->{TIMEOUT_HUNTER} < $NOW && int(keys(%{$tref->{PIECE_INWORK}})) < MAX_PREREQUEST_PIECES) {
				# -> I am NOT choked an free.. create a request
				my $suggested_piece = Bitflu::Store::SuggestNewPiece($tref);
				l(L_DEBUG_SUGST, "<$sock> Attemping to request piece '$suggested_piece\@$tref->{INFO_HASH}' from client");
				
				if(defined($suggested_piece)) {
					# -> Request this piece from client
					my ($sug_off, $sug_size) = Bitflu::Store::GetFreePieceOffset($tref->{INFO_HASH},$suggested_piece);
					_xwA($sock, Bitflu::Protocol::CreateRequest({index=>$suggested_piece, 
					      offset=> $sug_off, size=>$sug_size}));
					Bitflu::Store::SetPieceAsInwork($tref->{INFO_HASH},$suggested_piece);
					$tref->{PIECE_INWORK}{$suggested_piece} = $suggested_piece;
					$tref->{TIMEOUT_WAITFORPIECE} = $NOW + TIMEOUT_WAITFORPIECE;
				}
				else {
					$tref->{TIMEOUT_HUNTER} = $NOW + TIMEOUT_HUNTER;
					l(L_INFO, "Hunter: <$sock> has no new pieces for me");
					
					if(int(keys(%{$tref->{PIECE_INWORK}})) == 0) {
						l(L_INFO, " .. and no unfullfilled requests. Choking myself");
						_xwA($sock, Bitflu::Protocol::CreateNotInterested());
						$tref->{ME_CHOKE} = 1;
						$tref->{TIMEOUT_MECHOKED} = $NOW + TIMEOUT_MECHOKED;
						FreeWorkPieces($tref);
					}
					
				}
			}
			
		} #########################################################
		else {
			# First checkout if this non-idle client timed out
			if($tref->{TIMEOUT} < $NOW) {
				l(L_INFO, "<$sock> timed out in state <$state>");
				killSock($torrent_handle,$sock);
				next;
			}
			
			# Still here? Ok, lets check if we got something to send
			if($state == BTS_WRITE_HANDSHAKE) {
				l(L_DEBUG, "<$sock> -> Performing Handshake");
				_xwA($sock, Bitflu::Protocol::CreateHandshake({info_hash=>$tref->{INFO_HASH},
				                                                  peer_id=>$config->{client}->{peerid}}));
				$tref->{BTS}    = BTS_READ_HANDSHAKE_RESPONSE;
				$tref->{TIMEOUT} = $NOW+TIMEOUT_SHAKE;
			}
			elsif($state == BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD) {
				l(L_DEBUG, "<$sock> -> Sending Handshake");
				_xwA($sock, Bitflu::Protocol::CreateHandshake({info_hash=>$tref->{INFO_HASH},
				                                                  peer_id=>$config->{client}->{peerid}}));
				$tref->{BTS} = BTS_WRITE_BITFIELD;
				$tref->{TIMEOUT} = $NOW+TIMEOUT_SHAKE;
			}
			elsif($state == BTS_WRITE_BITFIELD) {
				l(L_INFO, "<$sock> -> Writing Bitfield and entering IDLE state");
				_xwA($sock, Bitflu::Protocol::CreateBitfield($TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{bitfield}));
				$tref->{BTS} = BTS_IDLE;
				delete($tref->{TIMEOUT}); #The Handshake timeout is not checked in BTS_IDLE mode.. freeup some memroy..
				$tref->{SENT_KEEPALIVE} = $NOW + 5; #This is WRONG but there are some broken clients so we won't wait a full tick to send the keepalive
				$tref->{RECEIVED_KEEPALIVE} = $NOW + TIMEOUT_KEEPALIVE;  #Wait 'till X for a keepalive
				$tref->{TIMEOUT_MECHOKED}   = $NOW + TIMEOUT_MECHOKED;   #Drop connection if we are choked for more than X seconds
				# => TIMEOUT_WAITFORPIECE not set because PIECE_INWORK is set to UNDEF
				# => TIMEOUT_INTERESTED   not set because we should send INTERESTED packages ASAP
				# => TIMEOUT_HUNTER       not set because hunt should start ASAP
			}
			elsif($state == BTS_KILLME_PLEASE) {
				l(L_DEBUG, "<$sock> killing client as requested");
				killSock($torrent_handle,$sock);
				next;
			}
		}
		
		
		
		# -> WRITE data to wire
		my $buff = substr(_xGETtowrite($sock),0,$upcredits);
		my $bufflen = length($buff);
		if($bufflen) {
			___xxx_ugly_genbwstats($NOW);
			if($BWSTATS{UPX} > $config->{client}->{maxup}) {
				l(L_DEBUG,"=> Running too fast! ($upcredits) !!");
				$buff = substr($buff,0,8);
			}
			my $data_sent = syswrite($sock,$buff);
			$BWSTATS{UP}+=$data_sent;
			l(L_DEBUG, "<$sock> -> SENT $data_sent bytes to client");
			_xCLEANtowrite($sock,$data_sent);
			
			if(!defined($data_sent)) {
				l(L_WARN, "<$sock> !! SENDING data failed with error $!");
				killSock($torrent_handle,$sock); #Close connection NOW
			}
		}
	}
	return ($clients);
}

sub FreeWorkPieces {
	my($tref) = @_;
	foreach my $workpiece (keys(%{$tref->{PIECE_INWORK}})) {
		Bitflu::Store::SetPieceAsFree($tref->{INFO_HASH}, $workpiece);
	}
	delete($tref->{PIECE_INWORK});
}

#################################################################
# Returns the number of unchoked peers
sub GetNumberOfUnchokedPeers {
	my $c = 0;
	foreach my $xs (keys(%TCSTAT)) {
		$c++ if ($TCSTAT{$xs}->{PEER_CHOKE} == 0);
	}
	return $c;
}







package Bitflu::HTTPgui;
use URI::Escape; # Imported in main:: but we would like to get the exported vars here :-)

sub main {
	my($rq) = @_;
	
	my @HTTP_HEADER = split(/\r\n/,$rq);
	my $http_buff = undef;
	my($query_method, $uri) = $HTTP_HEADER[0] =~ /^(GET|HEAD) (\S+) HTTP\//i;
	
	my($rq_hash, $rq_cmd) = $uri =~ /^\/([^\/]+)\/(\S+)$/;
	
	unless($query_method) {
		return "HTTP/1.0 500 Invalid Request\r\n\r\nInvalid Request\r\n";
	}
	
	
	$http_buff = "HTTP/1.0 200 OK\r\nServer: Bitflu\r\n";
	
	
	if($rq_hash eq "Images") {
		$http_buff .= "Content-Type: image/gif\r\n\r\n";
		$http_buff .= CreateGifImage(split(/_/,$rq_cmd));
		return $http_buff;
	}
	elsif($rq_hash eq "StyleSheet") {
		$http_buff .= "Content-Type: text/css\r\n\r\n";
		$http_buff .= CreateCSS();
		return $http_buff;
	}
	else {
		$http_buff .= "Content-Type: text/html\r\n\r\n";
	}
	# Normal HTTP-View
	
	
	
	$http_buff .= "<html><header><title>Bitflu $rq_hash</title>\n";
	$http_buff .= '<link rel="stylesheet" href="/StyleSheet/fark.css" type="text/css">';
	$http_buff .= "</header><body>\n<h3>Bitflu HTTP-Interface</h3>\n";
	
	$http_buff .= "<form method=get action='/Load/'><input type=text name=file><input type=submit value='Load torrent'></form>\n";
	
	my $injected = undef;
	
	if($rq_hash eq "Load") {
		$rq_cmd = substr($rq_cmd,1); #Remove ?
		my $toload   = undef;
		my $takethis = 0;
		foreach my $xpart (split(/&/,$rq_cmd)) {
			foreach my $xm (split(/=/,$xpart)) {
				next if defined($toload);
				$toload = uri_unescape($xm) if $takethis;
				$takethis=1 if $xm eq 'file';
			}
		}
		$toload =~ tr/ //d;
		$http_buff .= BuildDialogWindow("Loading ".uri_escape($toload));
		Bitflu::RPC::execute("load $toload");
	}
	elsif(!$TORRENT_UNIVERSE->{$rq_hash}) {
		$http_buff .= "<!-- No valid hash //-->\n";
	}
	elsif($rq_cmd eq "ClientStats") {
		$injected = BuildDialogWindow(BuildClientStats($rq_hash));
	}
	elsif($rq_cmd eq "Cancel") {
		$injected = BuildDialogWindow("Do you really want to cancel $rq_hash ?", [{ 'Yes' => "/$rq_hash/RealCancel"}, {'Abort' => "/$rq_hash/Selected"}]);
	}
	elsif($rq_cmd eq "RealCancel") {
		Bitflu::RPC::execute("cancel $rq_hash");
		$http_buff .= "<script language=\"JavaScript\">this.location = '/';</script>";
	}
	elsif($rq_cmd eq "Commit") {
		$injected = BuildDialogWindow("Start commiting $rq_hash ?", [ { 'Start Commit' => "/$rq_hash/RealCommit"} , {'Abort' => "/$rq_hash/Selected"}]);
	}
	elsif($rq_cmd eq "RealCommit") {
		Bitflu::RPC::execute("commit $rq_hash");
		$injected = BuildDialogWindow("Commit of $rq_hash is running");
	}

	
	$http_buff .= BuildViewDownloads($rq_hash,$injected);
	
	$http_buff .= "</body></html>\n";
	return $http_buff;
}

sub BuildDialogWindow {
	my($txt, $aref) = @_;
	
	my $buff = "\n<div class='msgbox'><div>$txt</div>\n";
	
	my $buttons = undef;
	foreach my $ref (@$aref) {
		foreach my $k (keys(%$ref)) {
			$buttons .= "<a class='linkbox' href='$ref->{$k}'>$k</a>";
		}
	}
	
	$buff .= "<br>$buttons<br><font size-2> " if $buttons;
	
	$buff .= "</div>\n";
	return $buff;
}


##################################################################################################################################
# Show Commandlinks
sub BuildCommandList {
	my($hash) = @_;
	return BuildDialogWindow("<i>==&gt;$hash</i>", [ { "Commit" => "/$hash/Commit" },
	                                                 { "Cancel" => "/$hash/Cancel" },
	                                                 { "Show statistics" => "/$hash/ClientStats" },
	                                                 { "Close" => "/" } ]);
}

##################################################################################################################################
# Shows verbose information about a torrent
sub BuildClientStats {
	my($hash) = @_;
	
	my $http_buff = "<table border=1><tr><td>PeerID</td><td>IP</td><td>MeChoked</td><td>PeerChoked</td><td>Downloading</td><td>Ranking</td><td>Uploaded</td></tr>\n";
	
	foreach my $cs (keys(%TCSTAT)) {
		next if lc($TCSTAT{$cs}->{INFO_HASH}) ne lc($hash);
		my $px = undef;
		foreach(keys(%{$TCSTAT{$cs}->{PIECE_INWORK}})) { $px .= "$_;"; }
		$http_buff .= "<tr>";
		$http_buff .= "<td>"._htmlIze(_prfy($TCSTAT{$cs}->{CLIENT_ID}))."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{IP}."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{ME_CHOKE}."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{PEER_CHOKE}."</td>";
		$http_buff .= "<td>$px</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{RANKING}."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{STATS_UPLOADED}." X</td>";
		$http_buff .= "</tr>\n";
	}
	$http_buff .= "</table>\n";
	return $http_buff;
}


##################################################################################################################################
# Displays an HTML-Version of 'vd'
sub BuildViewDownloads {
	my($injectAt, $injected) = @_;
	my $http_buff = undef;
	$http_buff .= "<div class='xdesc'>Upload: ".sprintf("%.3f",$BWSTATS{UPX})." KiB/s &nbsp;&nbsp; Download: ".sprintf("%.3f",$BWSTATS{DOWNX})." KiB/s</div>\n<table style='border:thin solid black; border-spacing:0px'>\n";
	$http_buff .= "<tr><td>Name</td><td>Hash</td><td>Peers (Active/Total)</td><td>Pieces (Done/Total)</td><td>Size (Done/Total)</td><td>Done %</td></tr>\n";
			my %list = ();
			
			foreach my $s (keys(%TCSTAT)) {
				my $ih = $TCSTAT{$s}->{INFO_HASH};
				my $mc = $TCSTAT{$s}->{ME_CHOKE};
				$list{$ih}->{peers}++;
				$list{$ih}->{active}++ if $mc == 0;
			}
	my $dlcount = 0;
	foreach my $k (sort keys(%{$TORRENT_UNIVERSE})) {
		my $donestats = Bitflu::Store::GetDoneStats($k);
		my $blunt = "table1";
		   $blunt = "table2" if (++$dlcount%2);
		   $blunt = "selected" if lc($k) eq lc($injectAt);
		my $torrentstate = 'idle';
		   $torrentstate = 'downloading' if $list{$k}->{active} > 0;
		   $torrentstate = 'stalled'     if $list{$k}->{peers}  < 1;
		   $torrentstate = 'completed'   if $donestats->{pieces} == int(Bitflu::Store::GetNumberOfPieces($k)+1);
		
		my $percent_done = sprintf("%6.2f",(($donestats->{pieces}/(Bitflu::Store::GetNumberOfPieces($k)+1))*100));
		
		$http_buff .= "<tr class=$blunt>\n";
		$http_buff .= "<td class='$torrentstate'><b>"._htmlIze(substr(${$TORRENT_UNIVERSE->{$k}->{torrent}}->{content}->{info}->{name},0,96))."</b></td>";
		$http_buff .= "<td><a href='/$k/Selected'>$k</a></td>";
		$http_buff .= "<td>".int($list{$k}->{active})."/".int($list{$k}->{peers})."</td>";
		$http_buff .= "<td>".int($donestats->{pieces})."/".int(Bitflu::Store::GetNumberOfPieces($k)+1)."</td>";
		$http_buff .= "<td>".int($donestats->{size}/1024/1024)."/".int($TORRENT_UNIVERSE->{$k}->{total_size}/1024/1024)."</td>";
		$http_buff .= "<td><img src='/Images/80_145_80' height=8 width=".int($percent_done)." alt='$percent_done% completed'>";
		$http_buff .= "<img src='/Images/145_95_80' height=8 width=".int(100-$percent_done)." alt='$percent_done% completed'></td>";
		$http_buff .= "</tr>\n";
		
		if(lc($k) eq lc($injectAt)) {
			$injected ||= BuildCommandList($k);
			$http_buff .= "<tr class=$blunt><td>$injected</td><td></td><td></td><td></td><td></td><td></td></tr>";
		}
		
	}
	$http_buff .= "</table>\n";
	
	
	
return $http_buff;
}


sub CreateGifImage {
	my($red, $green, $blue) = @_;
	my $buff =  pack("H*", "47494638376101000100800100");
	   $buff .= pack("c",$red);
	   $buff .= pack("c",$green);
	   $buff .= pack("c",$blue);
	   $buff .= pack("H*", "ffffff2c00000000010001000002024401003b");
	return $buff;
}

sub CreateCSS {
	my $buff = << "EOF";
html {
    color: #000000; 
    background: #fefefe;
    font-family: serif;
}
h1, h2, h3 {
    text-align: left; 
    font-family: sans-serif;
}
h1 {
    font-size: 150%;
    line-height: 150%;
}
h2 {
    font-size: 130%;
}
h3 {
    font-size: 100%;
}
.selected {
	background: #b5c0cd;
}
.completed {
	color: #969145;
}
.stalled {
	color: #913737;
}
.downloading {
	color: #497c47;
}
.idle {
	color: #2a2a35;
}
.table1 {
	background: #e8e8e8;
}
.table2 {
	background: #f0f0f0;
}
.msgbox {
	background: #dedede;
	padding: 4px;
	border-style: solid;
	border-width: 2px;
	border-color: #c6c3bd;
	text-align: center;
}

.linkbox {
	background: #dfdcdf;
	border-style: solid;
	border-width: 1px;
	margin-right: 4px;
	padding: 4px;
	border-color: #000000;
}
.linkbox:hover {
	background: #e3dfc3;
}
a:link, a:visited {
    color: #1b56ce;
    font-weight: bold;
}
a:link:hover, a:visited:hover {
    color: #d81e1e;
}
EOF
return $buff
}

sub _htmlIze {
	my($string) = @_;
	$string =~ s/&/&nbsp;/gm;
	$string =~ s/>/&gt;/gm;
	$string =~ s/</&lt;/gm;
	return $string;
}

sub _prfy {
	my($string) = @_;
	$string =~ tr/[a-z][A-Z][0-9]_\.& -//cd;
	return $string;
}

1;





##################################################################################################################################
package Bitflu::Publisher;

	my $UGLY_FILE_FIND_REF = ();
	use constant FAKESHA => 'X' x 20;
	use constant FAKESUM => FAKESHA.FAKESHA;
	
	
	#################################################################
	# Ehr.. main()
	# returns (torrent_hash, human_message) ; torrent_has is undef on error
	sub main {
		my($cmdline) = @_;
		
		my $fakesha = FAKESHA;
		my $fakesum = FAKESUM;
		my $ref = {};           # Dull
		my $torrent = {};       # Torrent Reference
		my $fatalerr = undef;
		
		Carp::confess("BUGBUG: fakesum defined while entering main?! someone forget to cleanup stuff ; panic!") if $TORRENT_UNIVERSE->{$fakesum};
		
		
		# 'Parse' arguments
		foreach(split(/\s+/,$cmdline)) {
			my($k,$v) = split(/=/,$_);
			push(@{$ref->{$k}},$v);
		}
		
		
		$TORRENT_UNIVERSE->{$fakesum}->{torrent} = \$torrent; # Glue $torrent <-> $TORRENT_UNIVERSE
		# ..and pollute our $torrent reference
		$torrent->{content}->{'creation date'} = int(time());                        # Time of creation
		$torrent->{content}->{'created by'}    = "BitFlu ".main::BITFLU_VERSION;     # Watchout for broken torrents! :-)
		$torrent->{content}->{info}->{pieces}  = $fakesha;                           # A fake Checksum ; needed to get INIT() happy
		$torrent->{content}->{info}->{name}    = ${$ref->{name}}[0];                 # Name
		$torrent->{content}->{info}->{'piece length'} = 512*1024;                    # Currently hardcoded to 512kib
		
		# -> Create multiTracker entry
		# See http://wiki.depthstrike.com/index.php/P2P:Protocol:Specifications:Multitracker
		my $trackers_count = 0;
		foreach(@{$ref->{announce}}) {
			my @multitrackers  = ();
			foreach(split(/#/,$_)) {
				$torrent->{content}->{'announce'} ||= $_;
				push(@multitrackers,$_);
				$trackers_count++;
			}
			push(@{$torrent->{content}->{'announce-list'}}, \@multitrackers);
		}
		# Hups: No backups? Well: Delete useless announce-list
		delete($torrent->{content}->{'announce-list'}) if $trackers_count < 2;
		delete($torrent->{content}->{announce})        unless $ref->{announce}; # -> trackerless torrent
		
		# Add DHT nodes: NOTE: Bitflu supports IPv6 DHT-Nodes ; Silly Mainline doesn't and may
		# crash if it eats a .torrent created by bitflu with IPv6 nodes!
		foreach(@{$ref->{nodes}}) {
			if($_ =~ /^([0-9a-fA-F\.:]+):(\d+)$/) {
				push(@{$torrent->{content}->{nodes}}, [$1, $2]);
			}
			else {
				$fatalerr .= "Invalid DHT-Node '$_' ";
			}
		}
		
		
		# Check if the Torrent is 'bootable' . Well: Bitflu could use
		# a torrent without announce and nodes without any problems.
		# But this check avoids the distribution of non-mainline-compatible torrents.
		if(!$torrent->{content}->{announce} && !$torrent->{content}->{nodes}) {
			$fatalerr .= "No 'announce' and/or 'nodes' keyword ; ";
		}
		
		
		#Walk directory tree and find all files
		$UGLY_FILE_FIND_REF = (); #Clean it
		File::Find::find( {wanted => \&HandlePublishItem, no_chdir=>1}, ( Bitflu::Store::__get_publishbase()) );
		
		if(ref($UGLY_FILE_FIND_REF->{files}) ne "ARRAY") {
			$fatalerr .= "No valid files inside 'publish' found ; ";
		}
		
		
		# Last stop before we start to publish the file: Check if someone
		# has set fatalerr:
		if(defined($fatalerr)) {
			delete($TORRENT_UNIVERSE->{$fakesum});
			return (undef, "- $fatalerr  Nothing published. See 'help publish' for more information");
		}
		
		
		
		# Init TorrentStorage for our fake-torrent..
		Bitflu::Store::InitTorrent($fakesum);
		# ..and slap it together
		my $slapt_result = slapit($UGLY_FILE_FIND_REF,$torrent);
		
		# What we have done so far:
		# We got a working copy of the content inside workdir/$fakesha
		# We got a messy mess inside $TORRENT_UNIVERSE->{$fakesum}
		# We got a bencoded torrent inside $encoded_torrent
		
		# No. 0 : Get rid of the $TORRENT_UNIVERSE pollution
		delete($TORRENT_UNIVERSE->{$fakesum});
		
		# No. 0.5 : Check if we could publish it..
		if(defined($slapt_result)) {
			return (undef,"Failed to publish content: '$slapt_result'");
		}
		
		# No. 1: Fixup our torrent hashref and decode the torrent
		$torrent->{content}->{info}->{files} = $UGLY_FILE_FIND_REF->{files};
		
		# .. and add a NULLBYTE to each path element
		# DO NOT USE THIS LIST ANYMORE WITH SYSCALLS!
		foreach(@{$torrent->{content}->{info}->{files}}) {
			my $uref = $_;
			foreach(@{$uref->{path}}) {
				$_ .= "_" if $_ =~ /^\d+$/; #Ugly hack: Add a something to num-only entries to force a string directory.
			}
		}
		
		# Check for single-file torrent:
		if(int(@{$torrent->{content}->{info}->{files}}) == 1) {
			# Wowies! Single file torrent.. follow the stupidity and
			# shift some stuff around
			$torrent->{content}->{info}->{name}   ||= ${${$torrent->{content}->{info}->{files}}[0]->{path}}[0];
			$torrent->{content}->{info}->{length} = ${$torrent->{content}->{info}->{files}}[0]->{length};
			delete($torrent->{content}->{info}->{files});
		}
		
		$torrent->{content}->{info}->{name} ||= "Unnamed torrent\@".int(time()); #Fallback name
		
		# No. 2 : Mumble $torrent to get correct values from bencode() :-(
		$torrent->{content}->{info}->{name} .= "_" if $torrent->{content}->{info}->{name} =~ /^\d+$/; #Ugly hack: Add something to get a string
		
		# No. 3 : Save the torrent somewhere
		my $encoded_torrent = Bitflu::Bencoding::encode($torrent->{content});
		my $savefile = Bitflu::Store::__get_publishbase()."/publishTorrent_$$"; # We are placing it there to avoid crashes at startup if thisone is corrupt
		open(PUBLISHED, ">$savefile") or Carp::confess("Unable to write $savefile : $!\n");
			print PUBLISHED $encoded_torrent;
		close(PUBLISHED);
		
		
		# No. 4 : TestLoad it to get a torrent_hash
		my $torrent_hash = Bitflu::Bencoding::torrent2hash($savefile)->{torrent_hash};
		   $torrent_hash or Carp::confess("PANIC: TestLoading torrent failed?!");
		
		# No. 5 : Rename the old dir
		rename(Bitflu::Store::_getbase($fakesum), Bitflu::Store::_getbase($torrent_hash))
		   or return (undef,"Publishing $torrent_hash failed: Torrent does exist in queue!");
		
		my $load_result = main::__load_torrent($savefile);
		unlink($savefile) or Carp::confess("Unable to unlink $savefile : $!\n");
		
		unless($load_result) {
			return (undef, "- Loading $torrent_hash somehow failed...");
		}
		else {
			#Fixup config
			$TORRENT_UNIVERSE->{$torrent_hash}->{config}->{downloaded} = $UGLY_FILE_FIND_REF->{size};
			$TORRENT_UNIVERSE->{$torrent_hash}->{config}->{uploaded}   = $UGLY_FILE_FIND_REF->{size};
			return ($torrent_hash, "+ Content of ".Bitflu::Store::__get_publishbase()." has been moved into torrent $torrent_hash and is shared");
		}
	}
	
	sub HandlePublishItem {
		my $tocrop = Bitflu::Store::__get_publishbase();
		my $item = $_;
		#print ">CALL $File::Find::dir | $_ | $File::Find::name\n";
		if(-f $item) {
			my @patharray = split(/\//,__torrentpath($item,$tocrop));
			my $size      = int(-s $item);
			if($size == 0) {
				main::l(main::L_WARN, "[Publish] Empty file '@patharray' skipped"); #Fixme: We can't handle empty files.. slapit chokes on it
				return undef;
			}
			$UGLY_FILE_FIND_REF->{size} += $size;
			push(@{$UGLY_FILE_FIND_REF->{files}}, { length=>$size, path=>\@patharray });
		}
	}
	
	#################################################################
	# Splitup and SHA-Hash content inside 'publish'-dir
	# Note: This sub changes the $torrent->{content}->{info}->{pieces} ref.
	sub slapit {
		my($findref,$torrent) = @_;
		
		my $slapbytes_left = $torrent->{content}->{info}->{'piece length'} or Carp::confess("Somehow we got no piece-length?!");
		my $slappiece      = 0;      #First piece
		my $slapbuff       = undef;  #Buffer for sysread data
		my $slapbuffconcat = undef;  #Buffer for a full piece
		my $slapsha        = undef;  #Buffer for SHA-Sums
		my $done_size      = 0;      #How much we had read
		my $total_size     = $findref->{size}; #..how much we need to read
		
		foreach(@{$findref->{files}}) {
			my $cref = $_;
			my $unixpath = Bitflu::Store::__get_publishbase().'/'.join('/',@{$cref->{path}});
			open(SLAP, $unixpath) or return "Unable to open $unixpath : $!";
				while($slapbytes_left) {
					my $bytes_read = sysread(SLAP,$slapbuff,$slapbytes_left);
					Carp::confess("FATAL: Reading from $unixpath failed: '$!'") unless defined $bytes_read;
					$slapbuffconcat .= $slapbuff;   #Concat read stuff to piecebuff
					$slapbytes_left -= $bytes_read; #..and dec. read bytes
					$done_size      += $bytes_read; #..and inc. total read count
					main::l(main::L_DEBUG_PBLSH, "READ: $bytes_read from $unixpath :: left $slapbytes_left ($done_size -> $total_size)");
					
					if(($slapbytes_left == 0 or $done_size == $total_size) and length($slapbuffconcat) != 0) {
						my $filepath      = Bitflu::Store::_getfreedir(FAKESUM)."/".int($slappiece);
						my $filepath_done = Bitflu::Store::_getdonedir(FAKESUM)."/".int($slappiece);
						open(PIECE, ">$filepath") or return "Unable to write to $filepath : $!";
							print PIECE $slapbuffconcat;
						close(PIECE);
						
						$slapsha .= Digest::SHA1::sha1($slapbuffconcat);
						rename($filepath,$filepath_done) or return "Unable to rename $filepath into $filepath_done : $!";
						main::l(main::L_WARN, "[Publish]: Piece $slappiece ($unixpath) written & hashed...");
						#Clean/Fix slapt's
						$slapbuffconcat = undef;
						$slapbytes_left = $torrent->{content}->{info}->{'piece length'};
						$slappiece++;
					}
					last if $bytes_read < 1; #EOF ; jump to next piece
				}
			close(SLAP);
		}
		$torrent->{content}->{info}->{pieces} = $slapsha; #Fixup fake-checksum
		return undef;
	}
	
	
	#################################################################
	# Crop's away the local path
	sub __torrentpath {
		my($value,$tocrop) = @_;
		my($out) = $value =~ /^$tocrop\/?(.*)$/;
		return $out;
	}
	
	
1;









##################################################################################################################################
package Bitflu::Telnet;

#################################################################
# Handle telnet commands FIXME: This needs some reworking
sub parseCMD {
	my($handle,$s,$line) = @_;
	
	foreach my $part (split(/\r?\n/,$line)) {
		if($part eq "!") {
			$part = $config->{SYS}->{lastcmd}
		}
		else {
			$config->{SYS}->{lastcmd} = $part;
		}
		
		my @ARGS = split(/\s+/,$part);
		my $c = $ARGS[0];
		
		
		if($c eq "quit" or $c eq "q") {
			main::killSock($handle,$s);
		}
		elsif($c eq "publish") {
			my($publish_hash, $publish_msg) = Bitflu::Publisher::main($part);
			main::_xwA($s,$publish_msg."\n");
			Bitflu::RPC::execute("seed $publish_hash") if $publish_hash; #Set SHARE flag
		}
		elsif($c eq "load") {
			
			my $torrent_to_load   = $ARGS[1];
			my $torrent_to_delete = undef;
			## XXX HACK
			if($ARGS[1] =~ /^http:\/\//i) {
				my $uriget = Bitflu::Tracker::doHTTP({full_uri=>$ARGS[1]});
				if($uriget->{CODE} != 200) {
					main::_xwA($s, "- Unable to load $ARGS[1] ; Server responsed with code $uriget->{CODE}");
					next;
				}
				else {
					my $savefile = Bitflu::Store::__get_torrentbase()."/httpLoad";
					open(HTTPLOAD,">".$savefile) or Carp::confess("Unable to write to $savefile : $!");
					print HTTPLOAD $uriget->{BODY};
					close(HTTPLOAD);
					$torrent_to_load = $torrent_to_delete = $savefile;
				}
			}
			## END HACK
			
			my $result = main::__load_torrent($torrent_to_load);
			if(defined $result) {
				main::_xwA($s, "+ $ARGS[1] with hash $result loaded\n");
			}
			else {
				main::_xwA($s, "- Loading '$ARGS[1]' failed.\n");
			}
			
			if(defined($torrent_to_delete)) {
				unlink($torrent_to_delete) or Carp::confess("Unablet to delete $torrent_to_delete : $!");
			}
			
		}
		elsif($c eq "seed" and $TORRENT_UNIVERSE->{$ARGS[1]}) {
			$TORRENT_UNIVERSE->{$ARGS[1]}->{config}->{seeding} = 1;
			main::_xwA($s, "$ARGS[1] is now in SUPPORT-SEEDING mode\n");
		}
		elsif($c eq "unseed" and $TORRENT_UNIVERSE->{$ARGS[1]}) {
			$TORRENT_UNIVERSE->{$ARGS[1]}->{config}->{seeding} = undef;
			main::_xwA($s, "$ARGS[1] reverted back to normal seeding\n");
		}
		elsif($c eq "analyze") {
			my $bla = Bitflu::Bencoding::torrent2hash($ARGS[1]);
			delete $bla->{content}->{info}->{pieces};
			main::_xwA($s, Data::Dumper::Dumper($bla));
		}
		elsif($c eq "qtracker" and $TORRENT_UNIVERSE->{$ARGS[1]}) {
			Bitflu::Tracker::Query($ARGS[1]);
			main::BT_getPeers($ARGS[1]);
		}
		elsif($c eq "bitmask") {
			foreach(sort keys(%{$TORRENT_UNIVERSE})) {
				main::_xwA($s,"Bitfield of $_\n");
				main::_xwA($s, "#".unpack("B*", $TORRENT_UNIVERSE->{$_}->{bitfield})."\n\n");
			}
		}
		
		elsif( ($c eq "commit" or $c eq "xcommit") and $TORRENT_UNIVERSE->{$ARGS[1]}) {
			my $pid = fork();
			my $hash = $ARGS[1];
			if(defined($pid) && $pid == 0) {
				#-> CHILD
				$SIG{INT} = $SIG{HUP} = undef; #Do NOT save config on HUP
				# CloseDown all connections
				my @xtelnet =$telnet_handle->can_write(main::DYNASLEEP_MIN);
				my @xtorrent=$torrent_handle->can_write(main::DYNASLEEP_MIN);
				#attemp to clean all connections held on this fork
				foreach(@xtelnet)  { close($_) if $_; }
				foreach(@xtorrent) { close($_) if $_; }
				
				Bitflu::RPC::wall("$hash: Commit started with pid $$");
				my $assemble_start  = time();
				my $assemble_result = Bitflu::Store::AssembleData($hash);
				my $assemble_end    = time();
				Bitflu::RPC::wall("$hash: Commit finished after ".int($assemble_end-$assemble_start)." seconds");
				
				print "DEBUGXX: $assemble_result\n";
				
				if($assemble_result) {
					Bitflu::RPC::wall("$hash: Commit errors: $assemble_result ; file not removed from queue");
				}
				elsif($c eq "commit") {
					Bitflu::RPC::execute("cancel $hash");
					Bitflu::RPC::wall("$hash: Removed from queue");
				}
				#else: xcommit
				
				print "[$$] reached EXIT\n";
				
				exit(0);
			}
			
		}
		elsif( $c eq "verify" and $TORRENT_UNIVERSE->{$ARGS[1]} ) {
			Bitflu::Store::VerifyStorageDir($ARGS[1]);
		}
		elsif( $c eq "wall") {
			$ARGS[0] = "[Bitflu message] :";
			my @xtelnet =$telnet_handle->can_write(main::DYNASLEEP_MIN);
			foreach(@xtelnet) {
				next if $_ eq $s;
				main::_xwA($_, "@ARGS\r\n");
			}
		}
		elsif($c eq "vd") {
			my %list = ();
			
			foreach my $s (keys(%TCSTAT)) {
				my $ih = $TCSTAT{$s}->{INFO_HASH};
				my $mc = $TCSTAT{$s}->{ME_CHOKE};
				$list{$ih}->{peers}++;
				$list{$ih}->{active}++ if $mc == 0;
			}
			my $buff = (sprintf(" *** UPLOAD : %.3f KiB/s | DOWNLOAD %.3f KiB/s\n", $BWSTATS{UPX} , $BWSTATS{DOWNX}));
			$buff   .= (sprintf("> %-32s |%-40s| || %-9s | %-9s| SIZE (DONE/TOTAL)  | DONE %   | RATION | FLAGS\n",
			                     'Name', '================= Hash =================', 'A/T>PEERS', 'DONE/TOTAL'));
			foreach my $k (sort keys(%{$TORRENT_UNIVERSE})) {
				my $donestats = Bitflu::Store::GetDoneStats($k);
				my $flags     = " ";
				   substr($flags,0,1,"S") if $TORRENT_UNIVERSE->{$k}->{config}->{seeding};
				
				my $t_buff .= sprintf("* %-32s (%-40s) || %4d/%4d | %4d/%4d | %5d/%5d (MB)   | %6.2f   | %6.2f | %s\n", 
							substr(${$TORRENT_UNIVERSE->{$k}->{torrent}}->{content}->{info}->{name},0,32),
							$k, $list{$k}->{active},$list{$k}->{peers}, $donestats->{pieces}, Bitflu::Store::GetNumberOfPieces($k)+1,
							int($donestats->{size}/1024/1024), int($TORRENT_UNIVERSE->{$k}->{total_size}/1024/1024),
							(($donestats->{pieces}/(Bitflu::Store::GetNumberOfPieces($k)+1))*100),
							( $TORRENT_UNIVERSE->{$k}->{config}->{uploaded}/($TORRENT_UNIVERSE->{$k}->{config}->{downloaded}+1) ), $flags
							);
				if($donestats->{pieces} == Bitflu::Store::GetNumberOfPieces($k)+1) {
					$buff .= Bitflu::Ansi::Bold($t_buff); #Completed -> Blue
				}
				elsif($list{$k}->{active} != 0) {
					$buff .= Bitflu::Ansi::Green($t_buff); #Downloading -> Green
				}
				elsif($list{$k}->{peers} == 0) {
					$buff .= Bitflu::Ansi::Red($t_buff); #No peers -> Red
				}
				else {
					$buff .= $t_buff;
				}
			}
			main::_xwA($s,$buff);
		}
		elsif($c eq "dhtboot" && $TORRENT_UNIVERSE->{$ARGS[1]}) {
			my $dhtdir = Bitflu::Store::_getdhtdir($ARGS[1]);
			my @XM = glob($dhtdir."/*"); #Fixme: use readdir
			my $credits = main::GetMaxConnections() - int(keys(%TCSTAT));
			   $credits = $config->{client}->{grabpeers} if $credits > $config->{client}->{grabpeers};
			foreach my $xfile (List::Util::shuffle(@XM)) {
				if($xfile =~ /\/?([\d\.]+):(\d+)$/) {
					main::BT_initConnection({port=>$2, ip=>$1, hash=>$ARGS[1]}) if --$credits > 1;
					unlink($xfile) or Carp::confess("Unable to unlink $xfile : $!");
				}
			}
		}
		elsif($c eq "cancel" && $TORRENT_UNIVERSE->{$ARGS[1]}) {
			my $hash = $ARGS[1];
			my $buff = undef;
			#Ok, this is'nt such an easy job..
			#1 : Mumble all clients
			foreach my $rs (keys(%TCSTAT)) {
				next unless $TCSTAT{$rs}->{INFO_HASH} eq $hash;
				$TCSTAT{$rs}->{BTS} = main::BTS_UNDEF;
				$TCSTAT{$rs}->{TIMEOUT} = 0; #Let the client timeout on next can_write run
				delete($TCSTAT{$rs}->{PIECE_INWORK});
				delete($TCSTAT{$rs}->{INFO_HASH});
			}
			## Kill the torrent
			Bitflu::Store::UnstoreTorrent($hash);
			## Ok ; all clients for this info_hash are now muted, kill the torrent_Universe...
			delete($TORRENT_UNIVERSE->{$hash});
			## ..and finaly rename the file..
			Bitflu::Store::CancelDirectory($hash);
			main::_xwA($s,$buff);
		}
		elsif($c eq "cs") {
			my $c = 0;
			my $match = $ARGS[1];
			foreach my $cc (keys(%TCSTAT)) {
				next unless $TCSTAT{$cc}->{INFO_HASH} =~ /$match/i;
				$c++;
				my $client = $TCSTAT{$cc}->{CLIENT_ID};
				$client =~ tr/a-zA-Z0-9:\.-//cd;
				my $px = "";
				foreach(keys(%{$TCSTAT{$cc}->{PIECE_INWORK}})) { $px .= "$_;"; }
				
				my $xcpieces = unpack("B*", $TCSTAT{$cc}->{BITFIELD});
				   $xcpieces =~ tr/1//cd; $xcpieces = length($xcpieces);
				main::_xwA($s, sprintf(">> %-33s -- HASH: %-40s | STATE: %03d | MC: %d | PC: %d | RANKING %04d | PIECE %-18s | P %04d | U %d | %s\n", $client,
				$TCSTAT{$cc}->{INFO_HASH}, $TCSTAT{$cc}->{BTS}, $TCSTAT{$cc}->{ME_CHOKE}, $TCSTAT{$cc}->{PEER_CHOKE}, $TCSTAT{$cc}->{RANKING},
				$px, $xcpieces, $TCSTAT{$cc}->{STATS_UPLOADED}, $TCSTAT{$cc}->{IP}));
			}
			main::_xwA($s, "\n -> $c clients out of ".main::GetMaxConnections()."\n");
		}
		elsif($c eq "version") {
			main::_xwA($s,"This is BitFlu ".main::BITFLU_VERSION." (".main::BITFLU_BUILD."); Running since ".int( (int(time())-($config->{SYS}->{boottime})) / 60 )." min. as '$0' with pid $$\n");
		}
		elsif($c eq "die") { kill(2,$$); }
		elsif($c eq "help" or $c eq "?") {
			main::_xwA($s, "\nBitFlu Help\n");
			main::_xwA($s, "------------\n");
			main::_xwA($s, "load /tmp/x.torrent        - Start download of /tmp/x.torrent\n");
			main::_xwA($s, "load http://example.com... - Grab .torrent and start download\n");
			main::_xwA($s, "commit \$hash               - Assemble download and cancel file if completed\n");
			main::_xwA($s, "xcommit \$hash              - Same as 'commit' but does not remove file from queue (= still seeded)\n");
			main::_xwA($s, "vd                         - Printout download queue / stats\n");
			main::_xwA($s, "cancel \$hash               - Remove download with \$hash from download queue\n");
			main::_xwA($s, "seed \$hash                 - Do not penalty leeching clients\n");
			main::_xwA($s, "unseed \$hash               - Do penalty leeching clients (default)\n");
			main::_xwA($s, "publish announce=http://... - Create a torrent and seed it; Type '$c publish' for more information\n");
			main::_xwA($s, "version                    - Printout Version (".main::BITFLU_VERSION.") and uptime\n");
			main::_xwA($s, "cs [\$hash]                 - Printout ClientStats (ADVANCED)\n");
			main::_xwA($s, "qtracker \$hash             - Force-Query tracker for torrent (ADVANCED)\n");
			main::_xwA($s, "wall \$message              - Broadcast message to all connected telnet clients (ADVANCED)\n");
			main::_xwA($s, "bitmask                    - Printout bitmask for all downloads (ADVANCED)\n");
			main::_xwA($s, "analyze /tmp/x.torrent     - Analyze torrent file (ADVANCED)\n");
			main::_xwA($s, "verify \$hash               - Verify StorageDir of \$hash ; Takes a LONG time and may cause BitFlu to exit (ADVANCED)\n");
			main::_xwA($s, "dhtboot \$hash              - Force connect to peers found via DistributedHashTable (ADVANCED)\n");
			main::_xwA($s, "die                        - Shutdown BitFlu\n");
			main::_xwA($s, "quit                       - Disconnect telnet client\n");
			main::_xwA($s, "\n");
			
			if($ARGS[1] eq "publish") {
				main::_xwA($s, Bitflu::Ansi::Bold("Creating a .torrent file\n------------------------\n"));
				main::_xwA($s, Bitflu::Ansi::Bold("Step #1: Place your content inside ".Bitflu::Store::__get_publishbase()."\n"));
				main::_xwA($s, Bitflu::Ansi::Bold("Step #2: Create the torrent file:\n # publish announce=http://tracker.example.com:6969/announce [name=description]\n"));
				main::_xwA($s, Bitflu::Ansi::Bold(" -> Use 'nodes' to create a DHT torrent:\n # publish nodes=127.0.0.1:6681 [nodes=127.0.0.2:6682]\n"));
				main::_xwA($s, Bitflu::Ansi::Bold("-NOTE-: 'name' is optional ; you can specify multiple trackers: 'announe=http://tracker1 announce=http://tracker2 ...'\n"));
				main::_xwA($s, Bitflu::Ansi::Bold("         LoadBalancing is also possible, use '#' to splitup balanced trackers: announce=http://tracker1#http://tracker2\n"));
				main::_xwA($s, Bitflu::Ansi::Bold("Step #3: Distribute the created torrent file: ".Bitflu::Store::__get_torrentbase()."/\$hash\n"));
				main::_xwA($s, Bitflu::Ansi::Bold("-NOTE-: You can delete the contents of ".Bitflu::Store::__get_publishbase()." after running the 'publish' command\n\n"));
			}
			
		}
		else {
			main::_xwA($s, "Command '$c' not understood or arguments missing ; type 'help' for command-syntax.\n");
		}
	}
}

1;

##################################################################################################################################
##################################################################################################################################
package Bitflu::Bitfield;

sub SetBit {
	my($bfref, $num) = @_;
	my $string = "0" x $num;
	   $string .= 1;
	$$bfref = $$bfref | pack("B*", $string);
}

sub GetBit {
	my($bfref, $num) = @_;
	my $bitfield = unpack("B*", "$$bfref");
#	Carp::cluck("BUGBUG ".length($bitfield)." < $num") if length($bitfield) < $num; # Not really a bug: just an incomplete bitfield
	return(substr(unpack("B*", "$$bfref"),$num,1));
}
1;


##################################################################################################################################
##################################################################################################################################
package Bitflu::Tracker;

use constant HTTP_REQUEST_TIMEOUT => 10;
use constant MAX_TRACKER_RETRY     => 3; #Try to contact the tracker 3 times (including redirects)

use List::Util 'shuffle';

#################################################################
# Query tracker.. FIXME: This needs some reworking
sub Query {
	my($info_hash) = @_;
	Carp::confess("No valid info hash") unless $info_hash;
	
	my $ref = ${$TORRENT_UNIVERSE->{$info_hash}->{torrent}}
	  or Carp::confess("No ref for this info hash :-(");
	
	Carp::confess("BUG: $ref->{torrent_hash} != $info_hash") if $ref->{torrent_hash} ne $info_hash;
	
	my @tracker_list = ($ref->{content}->{announce});
	
	# Torrent has an announce-list : Build tracker_list array
	if(ref($ref->{content}->{'announce-list'}) eq "ARRAY") {
		@tracker_list = ();
		foreach my $tracker_array ((@{$ref->{content}->{'announce-list'}})) {
			foreach my $xtracker (shuffle(@{$tracker_array})) {
				push(@tracker_list,$xtracker);
			}
		}
	}
	
	# .. and kick blacklisted trackers
	for(my $i=0;$i<int(@tracker_list);$i++) {
		if($config->{client}->{trackerblacklist} && $tracker_list[$i] =~ /$config->{client}->{trackerblacklist}/) {
			main::l(main::L_WARN, "[TrackerList] Removing blacklisted tracker '$tracker_list[$i]' from announce-list");
			$tracker_list[$i] = undef;
		}
	}
	
	
	foreach my $current_tracker (@tracker_list) {
		next if $current_tracker !~ /^http:\/\//i; # non-http entry?!
		
		my $httpH = {peer_id=>$config->{client}->{peerid}, key=>$config->{client}->{key},
		             stats_downloaded=>$TORRENT_UNIVERSE->{$info_hash}->{config}->{downloaded},
		             stats_uploaded=>$TORRENT_UNIVERSE->{$info_hash}->{config}->{uploaded},
		             stats_left=>($TORRENT_UNIVERSE->{$info_hash}->{total_size}-$TORRENT_UNIVERSE->{$info_hash}->{config}->{downloaded}),
		             port=>$config->{client}->{port}, info_hash=>$info_hash, uri=>$current_tracker};
		
		
		my $result = undef;
		
		for(0..MAX_TRACKER_RETRY) {
			$result = doHTTP($httpH);
			if($result->{CODE} == 301 or $result->{CODE} == 302) {
				# -> Redirect
				delete($httpH->{uri});
				$httpH->{full_uri} = $result->{HEADER}->{location}
			}
			elsif($result->{CODE} == 200) {
				last;
			}
		}
		
		
		if($result->{CODE} != 200) {
			main::l(main::L_WARN, "Unable to contact tracker $current_tracker for $info_hash ; last response: $result->{CODE}");
			next;
		}
		elsif(!$result->{BODY}) {
			main::l(main::L_WARN, "[TrackerList] Tracker $current_tracker for $info_hash returned an empty body");
			next;
		}
		elsif(! ($TORRENT_UNIVERSE->{$info_hash}->{announce} = Bitflu::Bencoding::decode($result->{BODY}) ) ) {
			main::l(main::L_WARN, "[TrackerList] Tracker $current_tracker for $info_hash returned a non-bencoded response");
			next;
		}
		elsif( ref($TORRENT_UNIVERSE->{$info_hash}->{announce}) ne "ARRAY" ) {
			# -> Compact response
			my $compact_list = $TORRENT_UNIVERSE->{$info_hash}->{announce}->{peers};
			my @compact_peer = ();
			for(my $i=0;$i<length($compact_list);$i+=6) {
				my $chunk = substr($compact_list, $i, 6);
				my $a    = unpack("C", substr($chunk,0,1));
				my $b    = unpack("C", substr($chunk,1,1));
				my $c    = unpack("C", substr($chunk,2,1));
				my $d    = unpack("C", substr($chunk,3,1));
				my $port = unpack("n", substr($chunk,4,2));
				my $ip = "$a.$b.$c.$d";
				push(@compact_peer, {ip=>$ip, port=>$port, peer_id=>"COMPACT"});
			}
			$TORRENT_UNIVERSE->{$info_hash}->{announce}->{peers} = \@compact_peer;
		}
		
		main::l(main::L_WARN, "[TrackerList] Tracker $current_tracker has been used for $info_hash");
		return 1;
	}
	
	main::l(main::L_WARN, "[TrackerList] All trackers for $info_hash are down");
	return 0;
	
}

sub _http_encode_info_hash {
	my($in) = @_;
	my $buff = undef;
	foreach my $k ((1..(length($in)/2))) {
		$buff .= "%".substr($in,(($k-1)*2),2);
	}
	return $buff;
}

sub doHTTP {
	my($h) = @_;
	
	foreach (qw(stats_left stats_uploaded stats_downloaded)) {
		if($h->{$_} < 1) {
			#stats_left has a good chance to be below zero if we
			#received even 1 bad chunk..
			main::l(main::L_INFO, "Fixing invalid size of $h->{$_} -> $_");
			$h->{$_} = 0;
		}
	}
	
	
	my $peer_id       = $h->{peer_id}; #Fixme: needs HTTP uri encoding
	my $key           = $h->{key};
	my $port          = int($h->{port});
	my $downloaded    = int($h->{stats_downloaded});
	my $uploaded      = int($h->{stats_uploaded});
	my $left          = int($h->{stats_left});
	my $URI           = $h->{uri};
	my $FULL_URI      = $h->{full_uri};
	my $torrent_hash  = _http_encode_info_hash($h->{info_hash});
	
	my ($tracker_host, $tracker_port, $tracker_base) = undef;
	my %ret = ();
	
	my $query = "GET ";
	
	if($FULL_URI) {
		($tracker_host,$tracker_port,$tracker_base) = $FULL_URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;
		$query .= "/$tracker_base HTTP/1.0\r\nHost: $tracker_host:$tracker_port\r\n\r\n";
	}
	else {
		#1: Decode tracker infos
		($tracker_host,$tracker_port,$tracker_base) = $URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;
		main::l(main::L_INFO, "Tracker($URI) decoded: HOST:$tracker_host / PORT:$tracker_port / BASE:$tracker_base");
		
		#2: Create the GET request
		$query .= "/$tracker_base?info_hash=$torrent_hash";
		$query .= "&peer_id=$peer_id&port=$port&uploaded=$uploaded&downloaded=$downloaded&left=$left&key=$key&compact=1";
		$query .= " HTTP/1.0\r\nHost: $tracker_host:$tracker_port\r\n\r\n";
	}
	
	return({ERROR=>"Unable to extract parts"}) unless $tracker_host;
	
	
	my $socket = IO::Socket::INET->new(PeerPort=>$tracker_port, PeerHost=>$tracker_host, Proto=>'tcp', Timeout=>2) or return({ERROR=>"$!"});
	
	alarm(HTTP_REQUEST_TIMEOUT);
	$SIG{ALRM} = sub { die "TIMEOUT\n"; };

	warn "announce h = " . ::Dumper($h);
	warn "announce host = $tracker_host:$tracker_port";
	warn "announce: $query";

	eval {
		print $socket $query;
		($ret{CODE}) = <$socket> =~ /^HTTP\/\d\.\d (\d{3})/;
		while(<$socket>) {
			chomp($_); chop($_);
			if($_ =~ /^([^:]+): (.+)$/) {
				$ret{HEADER}->{lc($1)} = $2;
			}
			last if length($_) == 0;
		}
		
		while(<$socket>) {
			$ret{BODY} .= $_;
		}
	};
	alarm(0);
	$SIG{ALRM} = undef;
	
	if($@) {
		return({ERROR=>"Timed out"});
	}
	else {
		return \%ret;
	}
}

1;

##################################################################################################################################
##################################################################################################################################
package Bitflu::Bencoding;

use constant DEBUG => 0;




	sub decode {
		my($string) = @_;
		my @chars = split(//,$string);
		return _decode(\@chars);
	}
	
	sub encode {
		my($ref) = @_;
		return _encode($ref);
	}
	
	
	
	sub _encode {
		my($ref) = @_;
		
		my $encoded = undef;
		
		if(ref($ref) eq "HASH") {
			$encoded .= "d";
			foreach(sort keys(%$ref)) {
				$encoded .= length($_).":".$_;
				$encoded .= _encode($ref->{$_});
			}
			$encoded .= "e";
		}
		elsif(ref($ref) eq "ARRAY") {
			$encoded .= "l";
			foreach(@$ref) {
				$encoded .= _encode($_);
			}
			$encoded .= "e";
		}
		elsif($ref =~ /^(\d+)$/) {
			$encoded .= "i$1e";
		}
		else {
			# -> String
			$encoded .= length($ref).":".$ref;
		}
		return $encoded;
	}
	
	
	sub _decode {
		my($ref) = @_;
		
		my $cx = shift(@$ref);
		if($cx eq 'd') {
			# -> A Dict : Cycle-Decode chunks until we get to an 'e'
			my $stolen_item = shift(@{$ref});
			my %ret_hash = ();
			while(defined($stolen_item) && $stolen_item ne 'e') {
				unshift(@{$ref},$stolen_item);
				my $key = _decode($ref);
				my $val = _decode($ref);
				$ret_hash{$key} = $val;
				print "DICT + $key -> $val\n" if DEBUG;
				$stolen_item = shift(@{$ref}); #NextOne
			}
			return \%ret_hash;
		}
		elsif($cx =~ /^\d$/) {
			# -> LengthPrefixed string
			print "STRING\n" if DEBUG;
			my $string = undef;
			my $length = $cx;
			
			my $stolen_item = shift(@{$ref});
			while(defined($stolen_item) && $stolen_item ne ':') {
				$length .= $stolen_item;
				$stolen_item = shift(@{$ref});
			}
			for(1..$length) {
				$string .= shift(@{$ref});
			}
			print "STRING + $string\n" if DEBUG;
			return $string;
		}
		elsif($cx =~ 'i') {
			# -> An int, cycle until we find a non-int (e)
			my $fullnum = undef;
			while(defined (my $append = shift(@{$ref}))) {
				last if $append !~ /^\d$/;
				
				$fullnum .= $append;
			}
			print "NUM: $fullnum\n" if DEBUG;
			return $fullnum;
		}
		elsif($cx eq 'l') {
			# -> A List (aka Array) ; Cycle-Decode until we got an 'e'
			my $stolen_item = shift(@{$ref});
			my @ret_array = ();
			while(defined($stolen_item) && $stolen_item ne 'e') {
				unshift(@{$ref}, $stolen_item);
				push(@ret_array, _decode($ref));
				$stolen_item = shift(@{$ref});
			}
			return \@ret_array;
		}
		elsif(defined($cx)) {
			warn "$0::Bitflu::Bencoding : Unexpected entry '$cx' found; Aborting parsing due to syntax error.\n";
			return undef;
		}
	}

#################################################################
# Load a torrent file
sub torrent2hash {
	my($file) = @_;
	my $buff = undef;
	open(BENC, $file) or return undef;
	while(<BENC>) {
		$buff .= $_;
	}
	close(BENC);
	return {} unless $buff;
	my $href = decode($buff);
	return {} unless ref($href) eq "HASH";
	my $torrent_hash =  Digest::SHA1::sha1_hex(encode(${$href}{info}));
	return {torrent_hash=>$torrent_hash, content=>$href};	
}



1;


##################################################################################################################################
##################################################################################################################################
package Bitflu::Protocol;

use constant BT_PIECE   => 7;
use constant BT_REQUEST => 6;
use constant BT_BITFIELD=> 5;
use constant BT_HAVE    => 4;
#################################################################
# Assembles a piece request
# Needs {index, offset, size}
sub CreateRequest {
	my($hr) = @_;
	Carp::confess("BUGBUG: Wrong CreateRequest call!") if ($hr->{offset} <0 or $hr->{index} < 0 or $hr->{size} < 1);
	main::l(main::L_DEBUG_CRX, "CreateRequest { PIECE : $hr->{index} | | OFFSET : $hr->{offset} | SIZE : $hr->{size} }");
	my $x = pack("N",13);
	$x .= pack("c", BT_REQUEST);
	$x .= pack("N", $hr->{index});
	$x .= pack("N", $hr->{offset});
	$x .= pack("N", $hr->{size});
	return $x;
}

sub CreateBitfield {
	my($bitfield) = @_;
	my $x = pack("N", (1+length($bitfield)));
	$x   .= pack("c", BT_BITFIELD);
	$x   .= $bitfield;
	return $x;
}

#################################################################
# Assembles a handshake request
# Needs {info_hash, peer_id}
sub CreateHandshake {
	my($hr) = @_;
	Carp::confess("BUGBUG") unless $hr->{info_hash} or $hr->{peer_id};

	my $extensions = undef;
	 $extensions .= pack("c7",0);
	 $extensions .= pack("B8", "00000001"); #Enables DHT

	my $x .= pack("c", 19);
	$x .= "BitTorrent protocol";
#	$x .= pack("c8",0);
	$x .= $extensions;
	$x .= pack("H40", $hr->{info_hash});
	$x .= $hr->{peer_id};
	return $x;
}

sub CreatePiece {
	my($hr) = @_;
	my $x .= pack("N", 9+$hr->{size});
	$x    .= pack("c", BT_PIECE);
	$x    .= pack("N", $hr->{index});
	$x    .= pack("N", $hr->{offset});
	$x    .= $hr->{data};
	return $x;
}

#################################################################
# Returns a Keepalive request
sub CreateKeepalive {
	return(pack("cccc",0000));
}

sub CreateHave {
	my $x = pack("N",5);
	$x   .= pack("c", BT_HAVE);
	$x   .= pack("N",$_[0]);
	return $x;
}

#################################################################
# Returns an Unchoke request
sub CreateUnchoke {
	my $x = pack("N",1);
	   $x .= pack("c",1);
	return $x;
}

#################################################################
# Returns a Choke request
sub CreateChoke {
	my $x = pack("N",1);
	   $x .= pack("c",0);
	return $x;
}

#################################################################
# returns an INTERESTED request
sub CreateInterested {
	my $x = pack("N",1);
	   $x .= pack("c",2);
	return $x;
}

#################################################################
# returns an NOT-INTERESTED request
sub CreateNotInterested {
	my $x = pack("N",1);
	   $x .= pack("c",3);
	return $x;
}

#################################################################
# Parse handshake response, returns unparsed bytes in _->{ol}
sub ParseHandshake {
	my($buff) = @_;
	if(length($buff) >= 68) {
		if(unpack("c",substr($buff,0,1)) == 19 &&
		   substr($buff,1,19) eq "BitTorrent protocol") {
			my $info_hash = substr($buff,28,20);
			my $client_id = substr($buff,48,20);
			return({ol=>substr($buff,68), info_hash=>unpack("H40",$info_hash), client_id=>$client_id});
		}
		else {
			return({ol=>$buff});
		}
	}
	else {
		return({ol=>$buff});
	}
}


1;

package Bitflu::RPC;

##################################################################################################################################
# Execute a 'telnet' command
sub execute {
	my($cmd) = @_;
	$cmd =~ tr/\n\r//d;
	
	my $sock = IO::Socket::INET->new(PeerHost=>($config->{client}->{telnetbind} || '127.0.0.1'), PeerPort=>$config->{client}->{telnet}, Proto=>'tcp', Timeout=>2)
																or return("connection failed: $!");
	print $sock $cmd."\r\n";
	close $sock if $sock;
	return undef;
}

##################################################################################################################################
# Execute the 'wall' command
sub wall {
	my($string) = @_;
	return execute("wall $string");
}

1;


##################################################################################################################################
##################################################################################################################################
package Bitflu::Store;
	
	use constant WORKDIR => ".working";
	use constant FREEDIR => ".free";
	use constant DONEDIR => ".done";
	use constant CONFDIR => ".settings";
	use constant DHTDIR  => ".dht";
	use constant DHTBOOT => ".dhtboot";
	use constant TORRENTS => "torrents";
	
	#################################################################
	# Init Bitflu::Store -> Creates the root directores
	sub _INIT {
		my $rootbase = __get_rootbase() or Carp::confess("BUGBUG: NO ROOTBASE");
		my $torrentbase = __get_torrentbase() or Carp::confess("BUGBUG : NO TORRENTBASE");
		my $publishbase = __get_publishbase() or Carp::confess("BUGBUG : NO PUBLISHBASE");
		my $autoloadbase= __get_autoloadbase() or Carp::confess("BUGBUG : NO AUTOLOADBASE");
		unless(-d $rootbase) {
			mkdir($rootbase) or Carp::confess("Unable to create rootbase '$rootbase' : $!");
		}
		
		chdir($rootbase) or Carp::confess("Unable to chdir($rootbase) : $!");
		
		unless(-d $torrentbase) {
			mkdir($torrentbase) or Carp::confess("Unable to create torrentbase '$torrentbase' : $!");
		}
		unless(-d $publishbase) {
			mkdir($publishbase) or Carp::confess("Unable to create publishbase '$publishbase' : $!");
		}
		unless(-d $autoloadbase) {
			mkdir($autoloadbase) or Carp::confess("Unable to create autoloadbase '$autoloadbase' :$!");
		}
	}
	
	#################################################################
	# Stors given torrent
	sub StoreTorrent {
		my($from,$to) = @_;
		$to ||= Carp::confess("BUGBUG");
		$to = __get_torrentbase()."/".$to;
		my @statA = stat($from);
		my @statB = stat($to);
		if($statA[1] == $statB[1]) {
			return $to; #<- same file!
		}
		my $buff = undef;
		open(IN, "$from") or Carp::confess("Unable to open $from : $!");
		open(OUT, ">$to") or Carp::confess("Unable to open $to : $!");
		while(<IN>) { print OUT $_; }
		close(IN);
		close(OUT);
		return $to;
	}
	
	#################################################################
	# Removes a torrent from store
	sub UnstoreTorrent {
		my($hash) = @_;
		_getbase($hash); #Assert check: Does the file exist?
		my $file = TORRENTS."/".$hash;
		unlink($file) or Carp::confess("Failed to unlink $file : $!");
	}
	
	#################################################################
	# Moves a basedir
	sub CancelDirectory {
		my($hash) = @_;
		my $dir = _getbase($hash);
		if(-d $dir) {
			for(0..255) {
				my $dest = $dir."_Cancel$_";
				next if -e $dest;
				rename($dir,$dest) or Carp::confess("Unable to rename $dir -> $dest : $!");
				last;
			}
		}
	}
	
	#################################################################
	# Stores DHT-Node inside .dhtboot directory
	sub AddDhtNode {
		my($hash, $ip, $port) = @_;
		my $bootdir =_getdhtbootabledir($hash);
		$port = int($port);
		return 0 unless $ip =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/; # Fixme: needs sub _isValidIP(v6) ?
		my $file = "$bootdir/$ip:$port";
		open(DHTX, ">$file") or Carp::confess("Unable to write to $file : $!"); close(DHTX);
		return 1;
	}
	
	#################################################################
	# Return torrent basedir
	sub __get_torrentbase {
		return TORRENTS;
	}
	
	sub __get_publishbase {
		return $config->{client}->{publishdir};
	}
	
	
	#################################################################
	# Return main-root basedir
	sub __get_rootbase {
		return $config->{client}->{rootdir};
	}
	
	#################################################################
	# Returns Torrent-Autoload directory
	sub __get_autoloadbase {
		return $config->{client}->{autoloaddir};
	}
	
	#################################################################
	# Get base of given hash
	sub _getbase {
		my($hash) = @_;
		Carp::confess("NO HASH") unless $hash;
		return $hash;
	}
	
	#################################################################
	# Get workdir of given hash
	sub _getworkdir {
		return _getbase($_[0]."/".WORKDIR);
	}

	#################################################################
	# Get freedir of given hash
	sub _getfreedir {
		return _getbase($_[0]."/".FREEDIR);
	}
	
	#################################################################
	# Get DistributedHashTablePeers dir
	sub _getdhtdir {
		return _getbase($_[0]."/".DHTDIR);
	}
	
	sub _getdhtbootabledir {
		return _getbase($_[0]."/".DHTBOOT);
	}

	#################################################################
	# Get donedir of given hash
	sub _getdonedir {
		return _getbase($_[0]."/".DONEDIR);
	}

	#################################################################
	# Get configdir of given hash
	sub _getconfdir {
		return _getbase($_[0]."/".CONFDIR);
	}

	#################################################################
	# Returns number of pieces of given info_hash (0 = 1 piece!)
	sub GetNumberOfPieces {
		my($hash) = @_;
		my $pieces = $TORRENT_UNIVERSE->{$hash}->{pieces};
		Carp::confess("No pieces?") unless defined $pieces;
		return $pieces;
	}
	
	#################################################################
	# Saves config-ref for given hash
	sub SaveConfig {
		my($hash, $ref) = @_;
		my $base = _getconfdir($hash);
		foreach my $setting (keys(%$ref)) {
			open(CONF, ">$base/$setting") or Carp::confess("Unable to write '$setting' : $!");
			print CONF $ref->{$setting};
			close(CONF);
		}
	}

	#################################################################
	# Returns the config-ref for given hash
	sub GetConfig {
		my($hash) = @_;
		my $base = _getconfdir($hash);
		my $ref = ();
		opendir(CONFDIR, $base) or return $ref;
		while(my $key = readdir(CONFDIR)) {
			my $fp = $base."/".$key;
			open(FILE, $fp) or next;
			while(<FILE>) {
				$ref->{$key} .= $_;
			}
			close(FILE);
		}
		closedir(CONFDIR);
		return $ref;
	}
	
	#################################################################
	# Returns downloaded pieces + total size : Note: this is just a 'guess'
	sub GetDoneStats {
		my($hash) = @_;
		my $numpieces = 0;
		my $size      = 0;
		
		my $bitfield = unpack("B*", $TORRENT_UNIVERSE->{$hash}->{bitfield});
		   $bitfield =~ tr/1//cd;
		   $numpieces = length($bitfield);
		   $size      = $numpieces*$TORRENT_UNIVERSE->{$hash}->{piece_len};
		return({pieces=>$numpieces, size=>$size});
	}

	#################################################################
	# Returns basename of torrent-file
	sub _get_torrent_basename {
		my($hash) = @_;
		Carp::confess("No valid hash") unless $TORRENT_UNIVERSE->{$hash};
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		my $conout = _cure_name($torrent->{content}->{info}->{name} or "OUT");
		return($conout);
	}

	##################################################################
	# Returns full-path basename of torrent-file
	sub _get_concat_name {
		my($hash) = @_;
		my $conout = _get_torrent_basename($hash);
		my $base   = _getbase($hash);
		return($base."/".$conout);
	}
	
	
	#################################################################
	# Removes evil chars from a filename.. this isn't perfect :-/
	sub _cure_name {
		my($string) = @_;
		$string =~ s/\.\.//;
		$string =~ tr/\/|//d;
		return $string;
	}

	#################################################################
	# Assemble final output file
	sub AssembleData {
		my($hash) = @_;
		my $outbase  = _get_concat_name($hash);                   # Multifile -> Output dir | Single file -> filename
		my $partbase = _getdonedir($hash);                        # .done directory
		my $torrent  = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};  # Hashref of torrent
		my $outbase  = _getbase($hash)."/"._cure_name(_get_torrent_basename($hash));
		my $plength  = $torrent->{content}->{info}->{'piece length'};   #Normal-Piece-Length
		my $pnum     = GetNumberOfPieces($hash);                        #How many pieces are there?
		my $offset   = 0;
		my $return   = undef;
		
		return("$outbase exists; aborting commit") if -d $outbase;
		
		mkdir($outbase) or Carp::confess("FATAL: Unable to create '$outbase' : $!");
		$plength ||= Carp::confess("Assert \$plength != 0 failed");
		
		# Extract ALL files from the torrent
		foreach my $fhash (@{$torrent->{content}->{info}->{files}}) {
			my $this_len = $fhash->{length};
			my @this_path = @{$fhash->{path}};  #Full path from torrent
			my $this_name = $outbase."/";       #Filename will start with this
			
			#Create all Subdirectories needed for this file:
			for(my $i=0; 1+$i < int(@this_path); $i++) {
				my $safepath = _cure_name($this_path[$i]);
				$this_name .= "$safepath/";
				print "mkdir($this_name)\n";
				unless(-d $this_name) {
					mkdir($this_name) or Carp::confess("Unable to create '$this_name' : $!");
				}
			}
			# ..and attach the filename to $this_name
			$this_name .= _cure_name($this_path[-1]);
			
			print "Assembling file : $this_name at $offset\n";
			my $offsetrange_start       = $offset;
			my $offsetrange_end         = $offset + $this_len;
			
			#Data starts at $piecerange_start with offset $piecerange_start_offset
			my $piecerange_start        = int($offsetrange_start/$plength);
			my $piecerange_start_offset = $offsetrange_start-($piecerange_start*$plength);
			
			#..and ends at $piecerange_end until byte $piecerange_end_offset
			my $piecerange_end          = int($offsetrange_end/$plength);
			my $piecerange_end_offset   = $offsetrange_end-($piecerange_end*$plength);
			
			print "FILE IS AT: $piecerange_start ($piecerange_start_offset) == TILL ==> $piecerange_end ($piecerange_end_offset)\n";
			
			open(OUTFILE, ">$this_name") or Carp::confess("Unable to open '$this_name' for writing: $!");
			
				for my $cpiece ($piecerange_start..$piecerange_end) {
					unless ( open(PART, "$partbase/$cpiece") ) {
						warn "$partbase/$cpiece is missing! Simulating it using /dev/zero\n";
						$return .= "[$cpiece was missing; file incomplete!]";
						open(PART, "/dev/zero") or Carp::confess("No /dev/zero to simulate missing pieces :-(");
					}
							#In each case: PART is now a useable filehandle
							my $cpiece_offset  = 0;
							   $cpiece_offset  = $piecerange_start_offset if $cpiece == $piecerange_start;
							my $cpiece_toread  = GetLengthOfPiece($hash,$cpiece)-$cpiece_offset;
							   $cpiece_toread  = $piecerange_end_offset-$cpiece_offset if $cpiece == $piecerange_end;
							my $cpiece_buff    = undef;
							seek(PART,$cpiece_offset,0);
							my $cpiece_didread = sysread(PART,$cpiece_buff,$cpiece_toread);
							print OUTFILE $cpiece_buff;
							
							print "WP: ($cpiece S:$piecerange_start / E:$piecerange_end) FROM: $cpiece_offset read $cpiece_toread ; Really got: $cpiece_didread\n";
							Carp::confess("BUMMER: You found a bug: $cpiece_didread != $cpiece_toread") if $cpiece_didread != $cpiece_toread;
					close(PART);
			}
			
			close(OUTFILE);
			
			$offset = $offsetrange_end;
		}
		
		#Wowies: Not a single file extracted? Just slam it together:
		if($offset == 0) {
			$outbase .= "/"._cure_name(_get_torrent_basename($hash));
			open(CONCAT, ">$outbase") or Carp::confess("Unable to open '$outbase' for writing: $!");
			foreach my $cpiece (0..$pnum) {
				unless( open(PART, "$partbase/$cpiece") ) {
					warn "Simulating piece $cpiece\n";
					$return .= "[$cpiece was missing; file incomplete!]";
					open(PART, "/dev/zero") or die "Unable to open /dev/zero :-( : $!";
				}
				
				my $buff = undef;
				my $toread = GetLengthOfPiece($hash,$cpiece);
				my $didread = sysread(PART,$buff,$toread);
				Carp::confess("ShortRead?! $toread != $didread") if $toread != $didread;
				print CONCAT $buff;
				print "DONE($cpiece\@$hash)\n";
				close(PART);
			}
			close(CONCAT);
		}
		
		return $return;
	}


	#################################################################
	# Create base layout of a workdir
	sub InitTorrent {
		my($hash) = @_;
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		Carp::confess("InitTorrent($hash) : hash is unknown") unless $torrent;
		Carp::confess("BugBug! InitTorrent called a 2nd time?!") if $TORRENT_UNIVERSE->{$hash}->{total_size};
		## Calculate total_size
		my $total_size = 0;
		if($torrent->{content}->{info}->{files}) {
			foreach(@{$torrent->{content}->{info}->{files}}) {
				my $h = $_;
				$total_size += $h->{length};
			}
		}
		else {
			$total_size = $torrent->{content}->{info}->{length};
		}
		$TORRENT_UNIVERSE->{$hash}->{total_size} = $total_size;
		$TORRENT_UNIVERSE->{$hash}->{config}->{downloaded} = 0;
		$TORRENT_UNIVERSE->{$hash}->{config}->{uploaded}   = 0;
		$TORRENT_UNIVERSE->{$hash}->{config}->{seeding}    = undef;
		
		## Now we are going to calculate the number of pieces,
		## -> THIS makes GetNumberOfPieces functional!
		my $unpacked_sha = length(unpack("H*", $torrent->{content}->{info}->{pieces}));
		Carp::confess("$hash is a br0ken torrent : $unpacked_sha < 40") if $unpacked_sha < 40;
		$TORRENT_UNIVERSE->{$hash}->{pieces}     = int($unpacked_sha/40)-1;
		
		# Calculate size of LAST piece
		my $allbutone = $total_size - (GetNumberOfPieces($hash)
		                               * $torrent->{content}->{info}->{'piece length'});
		
		$TORRENT_UNIVERSE->{$hash}->{lastpiece_len} = $allbutone;
		$TORRENT_UNIVERSE->{$hash}->{piece_len}     = $torrent->{content}->{info}->{'piece length'};
		
		## -> Init working directories
		if(-d _getbase($hash)) {
			main::l(main::L_INFO, "Starting Rollback for existing workdir: $hash ...");
			opendir(RBACK, _getworkdir($hash)) or die "Unable to open workdir for $hash : $!\n";
				while(defined (my $x = readdir(RBACK))) {
					next if substr($x, 0,1) eq ".";
					main::l(main::L_INFO, "  -> Rollback $x");
					SetPieceAsFree($hash,$x);
				}
			closedir(RBACK);
		}
		else {
			mkdir(_getbase($hash));             # Creates base directory
			mkdir(_getworkdir($hash));          # Creates .workdir
			mkdir(_getfreedir($hash));          # Creates .free
			mkdir(_getdonedir($hash));          # Creates .done
			mkdir(_getconfdir($hash));          # Creates .settings
			mkdir(_getdhtbootabledir($hash));   # Creates .dhtboot dir
			mkdir(_getdhtdir($hash)) unless $torrent->{content}->{info}->{private}; # Creates DHT-Bootdir UNLESS torrent was set to PRIVATE
			
			for(0..GetNumberOfPieces($hash)) {
				# Init FakeBucks
				my $dst = _getfreedir($hash)."/".int($_);
				open(DUMMY, ">$dst") or Carp::confess("Unable to write file $dst : $!");
				close(DUMMY);
			}
			
		}
		
		# (Re?)-Add dht nodes for torrent
		if(ref($torrent->{content}->{nodes}) eq "ARRAY") {
			main::l(main::L_WARN, "Adding nodes of $hash to DHT-Bootable list");
			foreach my $ar (@{$torrent->{content}->{nodes}}) {
				next unless ref($ar) eq "ARRAY";
				my $ip = $$ar[0];
				my $port = int($$ar[1]);
				Bitflu::Store::AddDhtNode($hash, $$ar[0], $$ar[1]) or main::l(main::L_WARN, "AddDhtNode($hash, $$ar[0], $$ar[1]) skipped.");
			}
		}
		
		main::l(main::L_WARN, "Updating bitfield for $hash ; this may take some seconds...");
		my $bitfield = GetBitfieldFromDisk($hash);
		$TORRENT_UNIVERSE->{$hash}->{bitfield} = $bitfield;
		main::l(main::L_WARN, "Init of '$hash' finished!");
		return 0;
	}
	
	#################################################################
	# Calculate bitfield from data stored on disk
	sub GetBitfieldFromDisk {
		my($hash) = @_;
		my $bitfield = "";
		foreach my $piece (0..GetNumberOfPieces($hash)) {
			if(-e _getdonedir($hash)."/".int($piece)) {
				$bitfield .= "1";
			}
			else {
				$bitfield .= "0";
			}
		}
		return pack("B*",$bitfield);
	}
	
	#################################################################
	# Check for 'lost' chunks due to OS-Crashes (fsck)
	sub VerifyStorageDir {
		my($hash) = @_;
		
		my $donedir = _getdonedir($hash);
		my $freedir = _getfreedir($hash);
		my $workdir = _getworkdir($hash);
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		
		my $HAD_TO_FIX = undef;
		
		foreach my $piece (0..GetNumberOfPieces($hash)) {
			if(-e $donedir."/".int($piece)) {
				open(DATA, $donedir."/".int($piece)) or Carp::confess("BUGBUG");
				if(VerifyPiece($torrent->{content}->{info}->{pieces},$piece, *DATA)) {
					main::l(main::L_WARN, "$piece\@$hash: Verify OKAY");
				}
				else {
					main::l(main::L_WARN, "$piece\@$hash: Corrupted, requeueing chunk as not-done");
					
					#Delete broken piece
					unlink($donedir."/".int($piece)) or Carp::confess("Unable to delete piece $piece : $!\n");
					#..and create a newone in freeDir
					open(NP, ">".$freedir."/".int($piece)) or Carp::confess("Unable to write $piece : $!\n");  print NP undef;  close(NP);
					$HAD_TO_FIX++;
				}
				close(DATA);
			}
			elsif(-e $workdir."/".int($piece)) {
				main::l(main::L_WARN, "$piece\@$hash: in transaction ; okay");
			}
			elsif(! (-e $freedir."/".int($piece)) ) {
				main::l(main::L_WARN, "$piece\@$hash: Piece got lost somehow, re-creating a newone...");
				open(NP, ">".$freedir."/".int($piece)) or Carp::confess("Unable to write $piece : $!\n");  print NP undef;  close(NP);
				$HAD_TO_FIX++;
			}
		}
		
		if($HAD_TO_FIX) {
			main::l(main::L_PANIC, "$hash: Fixed $HAD_TO_FIX error(s) for $hash ; Please restart $0 to rebuild the bitfield");
			Carp::confess("EXITING DUE TO CORRUPTED CHUNKS");
			die;die;die
		}
		else {
			main::l(main::L_WARN, "$hash: no corruption found :-)");
			return 0;
		}
	}
	
	
	#################################################################
	# SHA1-Verify piece : Returns '1' if okay
	sub VerifyPiece {
		my($shastring,$piece, $fh) = @_;
		my $torrent_sha = unpack("H40", substr($shastring,$piece*20,20));
		my $sha = Digest::SHA1->new();
		seek($fh,0,0);
		   $sha->addfile($fh);
		my $file_sha = $sha->hexdigest;
		if($file_sha eq $torrent_sha) {
			return 1;
		}
		else {
			return 0;
		}
	}
	
	#################################################################
	# Returns true if PIECE of HASH is in freelist (= Not done and not in-transition)
	sub PieceIsFree {
		my($hash,$piece) = @_;
		return (-e _getfreedir($hash)."/".int($piece));
	}
	
	#################################################################
	# Mark piece to work
	sub SetPieceAsInwork {
		my($hash,$piece) = @_;
		main::l(main::L_DEBUG, "SetPieceAsInwork($hash,$piece)");
		Carp::confess("$piece is NOT free.. i can't set it to WORK!") unless PieceIsFree($hash,$piece);
		rename(_getfreedir($hash)."/".int($piece), _getworkdir($hash)."/".int($piece)) or Carp::confess("rename failed : $!");
		Carp::confess("Wowies") if PieceIsFree($hash,$piece);
	}

	#################################################################
	# Move piece into .done folder
	sub SetPieceAsDone {
		my($hash,$piece) = @_;
		main::l(main::L_DEBUG, "SetPieceAsDone($hash,$piece)");
		my $from = _getworkdir($hash)."/".int($piece);
		my $to   = _getdonedir($hash)."/".int($piece);
		rename($from,$to) or Carp::confess("Unable to rename $from to $to : $!");
	}
	
	#################################################################
	# Zeroes-out a piece from the workqueue
	sub ZeroWorkPiece {
		my($hash,$piece) = @_;
		unless(-e _getworkdir($hash)."/".int($piece)) {
			main::l(main::L_WARN, "BUGCHECK! Someone attemped to zeroOut an unused piece!");
			main::l(main::L_WARN, " HASH: $hash | PIECE : $piece");
			main::l(main::L_WARN, "================= TCSTAT DUMP ===============");
			main::l(main::L_WARN, Data::Dumper::Dumper(\%TCSTAT)."\n\n\n\r\n");
			sleep(2);
			main::l(main::L_WARN, "Throwing CONFESS\n");
			Carp::confess("\n\nZeroWorkPiece == BFD EXTIED!\n\n");
		}
		open(KILL, ">"._getworkdir($hash)."/".int($piece)) or die;
		close(KILL);
	}
	
	#################################################################
	sub GetFreePieceOffset {
		my($hash,$piece) = @_;
		my $file = _getfreedir($hash)."/".int($piece);
		Carp::confess("GetFreePiece?!") unless -e $file;
		
		my $size  = (-s $file);
		my $fplen = GetLengthOfPiece($hash,$piece);
		my $plen  = main::PIECESIZE;
		
		if( ($fplen - $size) < main::PIECESIZE ) {
			$plen = ($fplen - $size);
		}
		main::l(main::L_DEBUG,"GetFreePieceOffset($hash,$piece) => $size | $plen");
		return ($size,$plen);
	}
	
	#################################################################
	# 'rollback' a piece from workdir to freedir
	sub SetPieceAsFree {
		my($hash,$piece) = @_;
		main::l(main::L_DEBUG, "SetPieceAsFree($hash,$piece)");
		unless(-e _getworkdir($hash)."/".int($piece)) {
			main::l(main::L_WARN, "BUGCHECK! Someone attemped to free an unused piece!");
			main::l(main::L_WARN, " HASH: $hash | PIECE : $piece");
			main::l(main::L_WARN, "================= TCSTAT DUMP ===============");
			main::l(main::L_WARN, Data::Dumper::Dumper(\%TCSTAT)."\n\n\n\r\n");
			sleep(2);
			main::l(main::L_WARN, "Throwing CONFESS\n");
			Carp::confess("\n\nSetPieceAsFreeBUG == BFD EXTIED!\n\n");
		}
		rename(_getworkdir($hash)."/".int($piece),_getfreedir($hash)."/".int($piece)) or die;
	}
	
	#################################################################
	# Suggest a new piece to request for this client reference
	sub SuggestNewPiece {
		my($client_ref, $exclude) = @_;
		main::l(main::L_DEBUG_SUGST, "SuggestNewPiece($client_ref->{INFO_HASH},$exclude)");
		
		if(!$client_ref->{BITFIELD}) { #No bitfield = no hits for sure
			main::l(main::L_DEBUG_SUGST, "Aborting search as client has no bitfield");
			return undef;
		}
		elsif(defined($client_ref->{SNP_CACHE}) && $client_ref->{SNP_CACHE} != $exclude 
		        && PieceIsFree($client_ref->{INFO_HASH},$client_ref->{SNP_CACHE})) {
			main::l(main::L_DEBUG_SUGST, "## CACHE HIT! $client_ref->{SNP_CACHE}");
			return $client_ref->{SNP_CACHE};
		}
		else {
			main::l(main::L_DEBUG_SUGST, "## CACHE MISS");
		}
		
		
		
		# Tactic 1 : Client EXCLUDED a piece, we are going to see if we could featch the NEXT piece
		if(defined($exclude)) {
			my $plusone = $exclude+1;
			if($plusone <= GetNumberOfPieces($client_ref->{INFO_HASH}) &&
			   Bitflu::Bitfield::GetBit(\$client_ref->{BITFIELD},$plusone) &&
			   PieceIsFree($client_ref->{INFO_HASH},$plusone)) {
				main::l(main::L_DEBUG_SUGST, "NextPiece for $client_ref->{INFO_HASH} $plusone");
				$client_ref->{SNP_CACHE} = $plusone;
				return $plusone;
			}
		}


=head
		############ FIXME: EXPERIMENTAL
		# Ok: NextOne failed and we are also got no random piece
		# .. most likely we are almos at the end of the download: Good idea to start some rare pieces now.
		my $xp_info_hash = $client_ref->{INFO_HASH};
		my $xp_lowest_av = 0;
		my $xp_lowest_px = -1;
		for my $xp_current_piece (0..GetNumberOfPieces($xp_info_hash)) {
			next unless Bitflu::Bitfield::GetBit(\$client_ref->{BITFIELD},$xp_current_piece); #Client doesn't have this piece.. skip it!
			next if     Bitflu::Bitfield::GetBit(\$TORRENT_UNIVERSE->{$xp_info_hash}->{bitfield}, $xp_current_piece); #Got it anyway..
			next if defined($exclude) && $xp_current_piece == $exclude;
			my $xp_x = 0;
			foreach my $xp_tck (keys(%TCSTAT)) {
				my $xp_ref = $TCSTAT{$xp_tck};
				next if $xp_ref->{INFO_HASH} ne $xp_info_hash;
				$xp_x += Bitflu::Bitfield::GetBit(\$xp_ref->{BITFIELD},$xp_current_piece);
			}
			
			next if $xp_x == 0; #Dead piece..
			# -> xp_x is > 0
			if(($xp_lowest_av > $xp_x) or ($xp_lowest_px < 0)) {
				print "->RF: PieceIsFree($xp_info_hash,$xp_current_piece)\n";
				next unless PieceIsFree($xp_info_hash,$xp_current_piece); #Not free.. so not suggestable
				$xp_lowest_px = $xp_current_piece;
				$xp_lowest_av = $xp_x;
			}
			last if $xp_lowest_av == 1; #LowestOfTheLow
			
		}
		
		
		if($xp_lowest_px > -1) {
			main::l(main::L_DEBUG_SUGST, "RarestFirst: $xp_lowest_px (sources: $xp_lowest_av)");
			return($xp_lowest_px);
		}
		
		###############################
=cut


		# Tactic 2 : Try to grab a random piece
		for(0..30) {
			my $sug = int(rand(GetNumberOfPieces($client_ref->{INFO_HASH})));
			if(   Bitflu::Bitfield::GetBit(\$client_ref->{BITFIELD},$sug) 
			   && Bitflu::Bitfield::GetBit(\$TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{bitfield},$sug) == 0
			   && PieceIsFree($client_ref->{INFO_HASH},$sug)) {
				main::l(main::L_DEBUG_SUGST, "RandomPiece for $client_ref->{INFO_HASH} $sug");
				$client_ref->{SNP_CACHE} = $sug;
				return $sug;
			}
		}

=head
		# Faster? 3th tactic
		for my $xp_cp (0..GetNumberOfPieces($client_ref->{INFO_HASH})) {
			next unless Bitflu::Bitfield::GetBit(\$client_ref->{BITFIELD},$xp_cp); #Client doesn't have this piece.. skip it!
			next if     Bitflu::Bitfield::GetBit(\$TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{bitfield}, $xp_cp); #Got it anyway..
			next if (defined($exclude) && int($xp_cp) == $exclude);
			
			if(PieceIsFree($client_ref->{INFO_HASH},$xp_cp)) {
				main::l(main::L_DEBUG_SUGST, "WalkedDownPiece2 for $client_ref->{INFO_HASH} : $xp_cp");
				$client_ref->{SNP_CACHE} = $xp_cp;
				return $xp_cp;
			}
		}
=cut
		
		# Tactic 3 : Walk down the full .free list
		# This isn't such a big performance hit: There ain't much pieces left if we are doing Tactic #3
		my $freelist = 0;
		opendir(FREELIST, _getfreedir($client_ref->{INFO_HASH})) or Carp::confess("Unable to open freedir");
		while(defined (my $x = readdir(FREELIST))) {
			next if substr($x, 0,1) eq ".";
			$freelist++;
			next if (defined($exclude) && int($x) == $exclude);
			next unless Bitflu::Bitfield::GetBit(\$client_ref->{BITFIELD},$x);
			closedir(FREELIST);
			main::l(main::L_DEBUG_SUGST, "WalkedDownPiece for $client_ref->{INFO_HASH} : $x");
			$client_ref->{SNP_CACHE} = $x;
			return $x;
		}
		closedir(FREELIST);

		#Idee Fixme: Nun haben wir ja 2 walkdowns.. beide sind je nach situation ungsntig:
		# Wir knnten einfach die Stats vom torrent angucken und z.B. bei ber 80% done randomPiece berspringen und den besseren
		# walkdown nehmen..

		main::l(main::L_DEBUG_SUGST, "found NOTHING for $client_ref->{INFO_HASH} ; Freelist has $freelist piece(s) left");
		
		## We Should PENALTY this guy!
		
		return undef;
	}
	
	
	
	
	
	#################################################################
	# Retruns length of Piece # in hash X
	sub GetLengthOfPiece {
		my($hash, $piece) = @_;
		my $num = 0;
		if($piece == GetNumberOfPieces($hash)) {
			$num = $TORRENT_UNIVERSE->{$hash}->{lastpiece_len};
		}
		else {
			$num = $TORRENT_UNIVERSE->{$hash}->{piece_len};
		}
		return $num;
	}
	
	
	
	
	
	#################################################################
	# Read data from disk
	sub ReadPiece {
		my($ref) = @_;
		Carp::confess("ReadPIece was called without a HASH!") unless $ref->{hash};
		my $torrent = ${$TORRENT_UNIVERSE->{$ref->{hash}}->{torrent}};
		Carp::confess("BUGBUG: Torrent $ref->{hash} not found in TORRENT_UNIVERSE") unless ref($torrent) eq "HASH";
		
		my $file = _getdonedir($ref->{hash})."/".int($ref->{index});
		my $buff = undef;
		open(DAT, $file) or return undef;
		seek(DAT, $ref->{offset},0);
		my $bytes_read = read(DAT,$buff,$ref->{size});
		close(DAT);
		main::l(main::L_DEBUG_RPICE, "ReadPiece($ref) : OFFSET:$ref->{offset}, REQUESTED: $ref->{size} , READ: $bytes_read");
		return ($buff,$bytes_read); #We don't care if sysread failed..
	}
	
	
	
	
	
	#################################################################
	# Store and Verify a piece
	#
	# The piece will *NOT* be anymore in .work IF $rollover is != 0 
	#
	sub StorePiece {
		my($ref) = @_;
		Carp::confess("StorePiece was called without a HASH!") unless $ref->{hash};
		my $torrent = ${$TORRENT_UNIVERSE->{$ref->{hash}}->{torrent}};
		Carp::confess("BUGBUG: Torrent $ref->{hash} not found in TORRENT_UNIVERSE") unless ref($torrent) eq "HASH";
		
		
		my $client_points = 0;                  #Will be set to the points the client earned for this piece
		my $next_offset   = 0;                  #Will be set to the next offset if not a rollover did NOT happen
		my $piece         = $ref->{index};      #Piece we are storing
		my $next_size     = main::PIECESIZE;    #Size of next request if a rollover did NOT happen
		my $rollover      = undef;              #Did a rollover happen?
		my $verified      = undef;              #Was the piece VERIFIED (= SHA OK)
		my $work_file     = _getworkdir($ref->{hash})."/".int($piece); # File we are writing to

		my $length_of_full_piece    = GetLengthOfPiece($ref->{hash},$piece);
		
		my $length_of_saved_data    = (-s $work_file);
		my $offset_of_current_piece = $ref->{offset};
		
		main::l(main::L_DEBUG_STORE, "STRORE PIECE:$piece | DATLEN: $length_of_saved_data | APPEND: ".length($ref->{data}));
		
		if( !(-e $work_file) ) {
			#$rollover = 1;
			#$next_offset = -1;
			#$client_points -= 20;
			# -> This is now FATAL ## Adrian 20061106 (FIXME : Remove comments)
			Carp::confess("Piece $work_file was *NOT* requested ; is NOT in workdir !");
		}
		elsif($offset_of_current_piece != $length_of_saved_data) {
			# -> Nope
			Carp::cluck("Received unuseable data for $piece\@$ref->{hash}: Got $offset_of_current_piece data but expected to get $length_of_saved_data stuff");
			$rollover = 1;       # Go to next piece
			$next_offset = -1;
			$client_points -= 10;
			ZeroWorkPiece($ref->{hash},$piece);  #Piece is BR0ken -> Wipe data..
			SetPieceAsFree($ref->{hash},$piece); #..and requeue it as .free
		}
		else {
			# -> Ok, we can now store this chunk
			open(DATA, ">>$work_file") or Carp::confess("Unable to write to $work_file : $!");
			print DATA $ref->{data} or Carp::confess("WRITE failed: $!");
			close(DATA); #Needed to flush all buffers
			$length_of_saved_data = $next_offset = (-s $work_file); # -> update ;-)

			open(DATA, "$work_file") or die;
			
			
			if($length_of_saved_data == $length_of_full_piece) {
				main::l(main::L_DEBUG_STORE, "STORE -> Piece is COMPLETE");

				# -> Complete piece!
				$rollover = 1; # Do a rollover, even if the chunk was garbage
				$next_offset = -1;
				
				my $verify_ok = VerifyPiece($torrent->{content}->{info}->{pieces},$piece, *DATA);
				
				if($verify_ok) {
					# -> Chunk was okay
					$client_points += 5;
					$verified = 1;
					SetPieceAsDone($ref->{hash},$piece);
				}
				else {
					#Failed to verify hash
					main::l(main::L_WARN, "STORE : Verification of $piece\@$ref->{hash} FAILED : Starting ROLLBACK");
					$client_points -= 30;
					ZeroWorkPiece($ref->{hash},$piece);  #Verification failed ; Remove the br0ken data from disc..
					SetPieceAsFree($ref->{hash},$piece); #..and requeue the piece
				}
			}
			elsif($length_of_saved_data > $length_of_full_piece) {
				#Should never happen
				Carp::confess("BUGBUG! Saved too much data in $work_file");
			}
			elsif( ($length_of_full_piece - $length_of_saved_data) < main::PIECESIZE ) {
				#May be the last piece..  # FIXME: Duplicate code: GetFreePieceOffset does almost the same
				$next_size = $length_of_full_piece - $length_of_saved_data;
				Carp::carp("Wowies: Adapting \$next_size to $next_size bytes");
			}
			close(DATA);
		}
		return({index=>$piece, size=>$next_size, offset=>$next_offset, rollover=>$rollover, verified=>$verified, points=>$client_points});
	}

1;

package Bitflu::Ansi;
use constant ANSI_ESC    => "\x1b[";
use constant ANSI_BOLD   => '1;';
use constant ANSI_BLACK  => '30m';
use constant ANSI_RED    => '31m';
use constant ANSI_GREEN  => '32m';
use constant ANSI_YELLOW => '33m';
use constant ANSI_BLUE   => '34m';
use constant ANSI_RSET   => '0m';


sub Green {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_GREEN.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Blue {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_BLUE.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Yellow {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_YELLOW.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Red {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_RED.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Bold {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC."40m".ANSI_ESC.ANSI_BOLD.ANSI_YELLOW.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub AnsiCure {
	my($s) = @_;
	my $badend = chop($s);
	if($badend ne "\n") { $s .= $badend ; $badend = undef }
	return($s,$badend);
}
1;


