#!/usr/bin/perl -w

use strict;
use IO::Socket::INET;
use Data::Dumper;
use URI::Escape;

use constant HTTP_REQUEST_TIMEOUT => 10;
use constant DEBUG => 0;

sub _http_encode_info_hash
{
	my ($in) = @_;
	my $buff = undef;
	foreach my $k ( ( 1 .. ( length($in) / 2 ) ) )
	{
		$buff .= "%" . substr( $in, ( ( $k - 1 ) * 2 ), 2 );
	}
	return $buff;
}

sub _decode
{
	my ($ref) = @_;

	my $cx = shift(@$ref);
	if ( $cx eq 'd' )
	{

		# -> A Dict : Cycle-Decode chunks until we get to an 'e'
		my $stolen_item = shift( @{$ref} );
		my %ret_hash    = ();
		while ( defined($stolen_item) && $stolen_item ne 'e' )
		{
			unshift( @{$ref}, $stolen_item );
			my $key = _decode($ref);
			my $val = _decode($ref);
			$ret_hash{$key} = $val;
			print "DICT + $key -> $val\n" if DEBUG;
			$stolen_item = shift( @{$ref} );    #NextOne
		}
		return \%ret_hash;
	}
	elsif ( $cx =~ /^\d$/ )
	{

		# -> LengthPrefixed string
		print "STRING\n" if DEBUG;
		my $string = undef;
		my $length = $cx;

		my $stolen_item = shift( @{$ref} );
		while ( defined($stolen_item) && $stolen_item ne ':' )
		{
			$length .= $stolen_item;
			$stolen_item = shift( @{$ref} );
		}
		for ( 1 .. $length )
		{
			$string .= shift( @{$ref} );
		}
		print "STRING + $string\n" if DEBUG;
		return $string;
	}
	elsif ( $cx =~ 'i' )
	{

		# -> An int, cycle until we find a non-int (e)
		my $fullnum = undef;
		while ( defined( my $append = shift( @{$ref} ) ) )
		{
			last if $append !~ /^\d$/;

			$fullnum .= $append;
		}
		print "NUM: $fullnum\n" if DEBUG;
		return $fullnum;
	}
	elsif ( $cx eq 'l' )
	{

		# -> A List (aka Array) ; Cycle-Decode until we got an 'e'
		my $stolen_item = shift( @{$ref} );
		my @ret_array   = ();
		while ( defined($stolen_item) && $stolen_item ne 'e' )
		{
			unshift( @{$ref}, $stolen_item );
			push( @ret_array, _decode($ref) );
			$stolen_item = shift( @{$ref} );
		}
		return \@ret_array;
	}
	elsif ( defined($cx) )
	{
		warn
"$0::Bitflu::Bencoding : Unexpected entry '$cx' found; Aborting parsing due to syntax error.\n";
		return undef;
	}
}

sub decode
{
	my ($string) = @_;
	my @chars = split( //, $string );
	return _decode( \@chars );
}

sub doHTTP
{
	my ($h) = @_;

	foreach (qw(stats_left stats_uploaded stats_downloaded))
	{
		unless ( defined $h->{$_} )
		{

			#stats_left has a good chance to be below zero if we
			#received even 1 bad chunk..
			warn "Fixing invalid value of $_";
			$h->{$_} = 0;
		}
	}

	my $peer_id      = uri_escape( $h->{peer_id} );
	my $port         = int( $h->{port} );
	my $downloaded   = int( $h->{stats_downloaded} );
	my $uploaded     = int( $h->{stats_uploaded} );
	my $left         = int( $h->{stats_left} );
	my $URI          = $h->{uri};
	my $FULL_URI     = $h->{full_uri};
	my $torrent_hash = _http_encode_info_hash( $h->{info_hash} );
	my $event        = $h->{event} || "";

	my ( $tracker_host, $tracker_port, $tracker_base ) = undef;
	my %ret = ();

	my $query = "GET ";

	if ($FULL_URI)
	{
		( $tracker_host, $tracker_port, $tracker_base ) =
		  $FULL_URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;
		$query .=
"/$tracker_base HTTP/1.0\r\nHost: $tracker_host:$tracker_port\r\n\r\n";
	}
	else
	{

		#1: Decode tracker infos
		( $tracker_host, $tracker_port, $tracker_base ) =
		  $URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;

		#2: Create the GET request
		$query .=
		    "/$tracker_base"
		  . ( index( $tracker_base, "?" ) != -1 ? "&" : "?" )
		  . "info_hash=$torrent_hash";
		$query .=
"&peer_id=$peer_id&port=$port&uploaded=$uploaded&downloaded=$downloaded&left=$left&compact=1";
		$query .= "&event=$event" if $event;
		$query .= " HTTP/1.0\r\nHost: $tracker_host:$tracker_port\r\n\r\n";
	}

	return ( { ERROR => "Unable to extract parts" } ) unless $tracker_host;

	my $socket = IO::Socket::INET->new(
										PeerPort => $tracker_port,
										PeerHost => $tracker_host,
										Proto    => 'tcp',
										Timeout  => 2
	  )
	  or return ( { ERROR => "$!" } );

	$SIG{ALRM} = sub { die "TIMEOUT\n"; };
	alarm(HTTP_REQUEST_TIMEOUT);

	warn "announce h = " . ::Dumper($h);
	warn "announce host = $tracker_host:$tracker_port";
	warn "announce: $query";

	eval {
		print $socket $query;

		my $data = <$socket>;
		print "-> $data";
		( $ret{CODE} ) = $data =~ /^HTTP\/\d\.\d (\d{3})/;

		while (<$socket>)
		{
			print "-> $_";
			s/\s+$//;
			if (/^([^:]+): (.+)$/)
			{
				$ret{HEADER}->{ lc($1) } = $2;
			}
			last if /^\s*$/;
		}

		$ret{BODY} = "";
		if ( $ret{HEADER}->{'content-length'} )
		{
			my $s = $ret{HEADER}->{'content-length'};
			while ($s)
			{
				my $data;
				$s -= $socket->read( $data, $s );
				$ret{BODY} .= $data;
			}
		}
		else
		{
			while (<$socket>)
			{
				$ret{BODY} .= $_;
			}
		}
	};

	delete $SIG{ALRM};
	alarm(0);

	if ($@)
	{
		return ( { ERROR => "Timed out" } );
	}
	else
	{
		return \%ret;
	}
}

my $usage =
"USAGE $0 <announce_uri> <port> <info_hash> <peer_id> <uploaded> <downloaded> <left> [event=started|completed|stopped]";

my $h;

$h->{uri}       = shift or die "No announce_uri param\n$usage";
$h->{port}      = shift or die "No port param\n$usage";
$h->{info_hash} = shift or die "No info_hash param\n$usage";

$h->{peer_id} = shift or die "No peer_id param\n$usage";
if ( length( $h->{peer_id} ) == 40 )
{
	$h->{peer_id} = unpack( "A20", pack( "H40", $h->{peer_id} ) );
}
length( $h->{peer_id} ) == 20 or die "len(peer_id) != 20\n$usage";

$h->{stats_downloaded} = shift;
defined $h->{stats_downloaded} or die "No downloaded param\n$usage";

$h->{stats_uploaded} = shift;
defined $h->{stats_uploaded} or die "No uploaded param\n$usage";

$h->{stats_left} = shift;
defined $h->{stats_left} or die "No left param\n$usage";

$h->{event} = shift || "";
     $h->{event} eq ""
  || $h->{event} eq "started"
  || $h->{event} eq "completed"
  || $h->{event} eq "stopped"
  or die "Incorrect event type\n$usage";

my $result = doHTTP($h);

if ( $result->{BODY} )
{
	my $decoded = decode($result->{BODY});
	$decoded->{peers} =~ tr/\x09\x0a\x0d\x20-\x7f/./c if $decoded->{peers};
	print "body=" . Dumper($decoded);
	$result->{BODY} =~ tr/\x09\x0a\x0d\x20-\x7f/./c;
}

print "result=" . Dumper($result);
